__author__ = 'Carrycat'
import re
import uuid
import unicodedata

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from .mixin_serializers import PasswordMixin, EmailMixin, PhoneNumPinMixin
from ..settings import app_settings

User = get_user_model()


class AbstractSignUpSerializer(PasswordMixin, EmailMixin, serializers.Serializer):
    username = serializers.CharField()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.error_messages.update({
            "username_taken": User._meta.get_field("username").error_messages[
                "unique"
            ],
            "email_taken": _("A user is already registered with this e-mail address."),
        })
        if not app_settings.UNIQUE_EMAIL:
            self.fields["email"].required = False
        if not app_settings.SIGNUP_USERNAME_REQUIRED:
            self.fields["username"].required = False

    def validate_username(self, username):
        try:
            for validator in app_settings.username_validators:
                validator(username)
        except ValidationError as e:
            raise serializers.ValidationError(
                e.message,
                code='ASU011'
            )
        if User.objects.filter(username=username).exists():
            raise serializers.ValidationError(
                self.error_messages['username_taken'],
                code='ASU001'
            )
        return username

    def validate(self, attrs):
        attrs = super().validate(attrs)
        return attrs


class DefaultSignUpSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('first_name', 'last_name',)


def base_serializers():
    custom_serializer = app_settings.SERIALIZERS['SIGNUP_SERIALIZER']
    base = [custom_serializer, AbstractSignUpSerializer]
    if app_settings.USE_PHONENUMBER_FIELD:
        base.insert(1, PhoneNumPinMixin)
    return tuple(base)


def base_fields():
    base_fields = ('email', 'password1', 'password2', 'username',)
    if app_settings.USE_PHONENUMBER_FIELD:
        base_fields += ('phone_number', 'pin')
    return base_fields


class SignUpSerializer(*base_serializers()):

    def populate_user(self):
        user = User()
        return user

    def save(self):
        user = self.populate_user()
        password = self.validated_data.pop("password", None)
        if password:
            user.set_password(password)
        else:
            user.set_unusable_password()
        for attr, value in self.validated_data.items():
            setattr(user, attr, value)

        self.populate_username(user)
        user.save()
        return user

    def populate_username(self, user):
        """
        Fills in a valid username, if required and missing.  If the
        username is already present it is assumed to be valid
        (unique).
        """
        first_name = user.first_name
        last_name = user.last_name
        email = user.email
        user.username = user.username or self.generate_unique_username([email, last_name, first_name])

    def generate_unique_username(self, txts):
        regex = r'[^\w\s@+.-]'
        username = None
        txts = list(filter(None, txts))
        for txt in txts:
            if not txt:
                continue
            txt = unicodedata.normalize('NFKD', force_str(txt))
            txt = txt.encode('ascii', 'ignore').decode('ascii')
            txt = force_str(re.sub(regex, '', txt).lower())
            txt = txt.split('@')[0]
            txt = txt.strip()
            txt = re.sub(r'\s+', '_', txt)
            try:
                username = self.validate_username(txt)
                break
            except serializers.ValidationError:
                pass
        if not username:
            username = ''.join(txts)
            if username == '':
                username = 'user'
                username += str(uuid.uuid4())[:4]
        return username

    class Meta:
        model = User
        fields = base_fields() + tuple(app_settings.SERIALIZERS['SIGNUP_SERIALIZER']().fields.keys())




from django.urls import re_path

from .views import FacebookLogin, GoogleLogin, AppleLogin, SocialConnections

urlpatterns = [
    re_path('^connections/$', SocialConnections.as_view(), name='social_connections'),
    re_path(r'^facebook/$', FacebookLogin.as_view(), name='fb_login'),
    re_path(r'^google/$', GoogleLogin.as_view(), name='gl_login'),
    re_path(r'^apple/$', AppleLogin.as_view(), name='apple_login'),
]


from .utils import render_mail
from celery import shared_task
from celery.utils.log import get_task_logger
logger = get_task_logger(name=__name__)


@shared_task
def send_email_task(email, template=None, subject='No title', **kwargs):
    if template:
        logger.info("Starting Send Email ", kwargs, template)
        msg = render_mail(subject, email, kwargs, template)
        msg.send()



import django


def pytest_configure(config):
    from django.conf import settings

    # USE_L10N is deprecated, and will be removed in Django 5.0.
    settings.configure(
        INSTALLED_APPS=[
            'django.contrib.auth',
            'django.contrib.sessions',
            'django.contrib.contenttypes',
            'rest_framework',
            'oauth2_provider',
            'auth_framework',
        ],
        DATABASES={
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:'
            },
        },
        REST_FRAMEWORK={
            'DEFAULT_AUTHENTICATION_CLASSES': (
                'oauth2_provider.contrib.rest_framework.OAuth2Authentication',
            ),
        },
        SECRET_KEY='key_for_auth_framework_tests',
        AUTHENTICATION_BACKENDS=[
            "auth_framework.backends.auth_backends.AuthenticationBackend",
        ],
        AUTH_USER_MODEL="auth_framework.User",
        TEMPLATES=[
            {
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'APP_DIRS': True,
            },
        ],
        OAUTH2_PROVIDER={
            "OIDC_ENABLED": True,
            "OIDC_RSA_PRIVATE_KEY": '''-----BEGIN RSA PRIVATE KEY-----
            MIIEogIBAAKCAQEAnzyis1ZjfNB0bBgKFMSvvkTtwlvBsaJq7S5wA+kzeVOVpVWw
            kWdVha4s38XM/pa/yr47av7+z3VTmvDRyAHcaT92whREFpLv9cj5lTeJSibyr/Mr
            m/YtjCZVWgaOYIhwrXwKLqPr/11inWsAkfIytvHWTxZYEcXLgAXFuUuaS3uF9gEi
            NQwzGTU1v0FqkqTBr4B8nW3HCN47XUu0t8Y0e+lf4s4OxQawWD79J9/5d3Ry0vbV
            3Am1FtGJiJvOwRsIfVChDpYStTcHTCMqtvWbV6L11BWkpzGXSW4Hv43qa+GSYOD2
            QU68Mb59oSk2OB+BtOLpJofmbGEGgvmwyCI9MwIDAQABAoIBACiARq2wkltjtcjs
            kFvZ7w1JAORHbEufEO1Eu27zOIlqbgyAcAl7q+/1bip4Z/x1IVES84/yTaM8p0go
            amMhvgry/mS8vNi1BN2SAZEnb/7xSxbflb70bX9RHLJqKnp5GZe2jexw+wyXlwaM
            +bclUCrh9e1ltH7IvUrRrQnFJfh+is1fRon9Co9Li0GwoN0x0byrrngU8Ak3Y6D9
            D8GjQA4Elm94ST3izJv8iCOLSDBmzsPsXfcCUZfmTfZ5DbUDMbMxRnSo3nQeoKGC
            0Lj9FkWcfmLcpGlSXTO+Ww1L7EGq+PT3NtRae1FZPwjddQ1/4V905kyQFLamAA5Y
            lSpE2wkCgYEAy1OPLQcZt4NQnQzPz2SBJqQN2P5u3vXl+zNVKP8w4eBv0vWuJJF+
            hkGNnSxXQrTkvDOIUddSKOzHHgSg4nY6K02ecyT0PPm/UZvtRpWrnBjcEVtHEJNp
            bU9pLD5iZ0J9sbzPU/LxPmuAP2Bs8JmTn6aFRspFrP7W0s1Nmk2jsm0CgYEAyH0X
            +jpoqxj4efZfkUrg5GbSEhf+dZglf0tTOA5bVg8IYwtmNk/pniLG/zI7c+GlTc9B
            BwfMr59EzBq/eFMI7+LgXaVUsM/sS4Ry+yeK6SJx/otIMWtDfqxsLD8CPMCRvecC
            2Pip4uSgrl0MOebl9XKp57GoaUWRWRHqwV4Y6h8CgYAZhI4mh4qZtnhKjY4TKDjx
            QYufXSdLAi9v3FxmvchDwOgn4L+PRVdMwDNms2bsL0m5uPn104EzM6w1vzz1zwKz
            5pTpPI0OjgWN13Tq8+PKvm/4Ga2MjgOgPWQkslulO/oMcXbPwWC3hcRdr9tcQtn9
            Imf9n2spL/6EDFId+Hp/7QKBgAqlWdiXsWckdE1Fn91/NGHsc8syKvjjk1onDcw0
            NvVi5vcba9oGdElJX3e9mxqUKMrw7msJJv1MX8LWyMQC5L6YNYHDfbPF1q5L4i8j
            8mRex97UVokJQRRA452V2vCO6S5ETgpnad36de3MUxHgCOX3qL382Qx9/THVmbma
            3YfRAoGAUxL/Eu5yvMK8SAt/dJK6FedngcM3JEFNplmtLYVLWhkIlNRGDwkg3I5K
            y18Ae9n7dHVueyslrb6weq7dTkYDi3iOYRW8HRkIQh06wEdbxt0shTzAJvvCQfrB
            jg/3747WSsf/zBTcHihTRBdAv6OmdhV4/dD5YBfLAkLrd+mX7iE=
            -----END RSA PRIVATE KEY-----''',
            'SCOPES': {
                "openid": "OpenID Connect scope",
                'read': 'Read scope',
                'write': 'Write scope',
            },
            'OAUTH2_VALIDATOR_CLASS': 'auth_framework.oauth.oauth2_validators.OauthValidator',
            'OAUTH2_BACKEND_CLASS': 'auth_framework.oauth.oauth2_backends.OAuthLibCore',
            # 'PKCE_REQUIRED': True,
        },
        ROOT_URLCONF='auth_framework.urls'
    )
    django.setup()



import os
import requests
from ..abstract_adapter import OauthAdapter

_GOOGLE_OAUTH2_CERTS_URL = "https://www.googleapis.com/oauth2/v1/certs"


class GoogleOauthAdapter(OauthAdapter):
    id = "google"
    access_token_url = "https://accounts.google.com/o/oauth2/token"
    profile_url = "https://www.googleapis.com/oauth2/v1/userinfo"
    public_key_url = "https://www.googleapis.com/oauth2/v3/certs"
    scope = ['email', 'profile']
    client_id = os.environ.get('GOOGLE_CLIENT_ID', '')
    client_secret = os.environ.get('GOOGLE_CLIENT_SECRET', '')

    def get_default_scope(self) -> list[str]:
        return self.scope

    def get_uid(self, data: dict[str:str]):
        access_token = data.get('access_token', None)
        if access_token:
            resp = requests.get(
                self.profile_url,
                params={"access_token": access_token, "alt": "json"},
            )
            resp.raise_for_status()
            data = resp.json()
            return data['id'], self.get_model_fields(data), data
        else:
            return super().get_uid(data)

    def get_model_fields(self, data:dict[str:str]) -> dict[str:str]:
        return dict(
            email=data.get("email"),
            last_name=data.get("family_name", ''),
            first_name=data.get("given_name", ''),
        )

    # def get_pubkey(self, token: str) -> RSAPublicKey:

    #     """
    #     * Fetches certificates from Google API.
    #     * generate JWT key file for decoding.
    #     """
    #     from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
    #     from cryptography.x509 import load_pem_x509_certificate
    #     from cryptography.hazmat.backends import default_backend
    #     jws = jwt.PyJWS()
    #     header = jws.get_unverified_header(token)
    #     key_id = header.get("kid")
    #     resp = requests.get(_GOOGLE_OAUTH2_CERTS_URL)
    #     resp.raise_for_status()
    #     try:
    #         certs_to_check = resp.json()[key_id]
    #         certs_to_check = str.encode(certs_to_check)
    #         cert_obj = load_pem_x509_certificate(certs_to_check, default_backend())
    #     except KeyError:
    #         raise InvalidPubKey("failed to get pubkey from Google with kid {kid}".format(kid=key_id))
    #     pub_key = cert_obj.public_key()
    #     return pub_key



import hashlib
import hmac
import os

import requests

from ..abstract_adapter import OauthAdapter

GRAPH_API_VERSION = 'v11.0'
GRAPH_API_URL = "https://graph.facebook.com/" + GRAPH_API_VERSION


class FacebookOauthAdapter(OauthAdapter):
    client_id = os.environ.get('FACEBOOK_CLIENT_ID', '')
    client_secret = os.environ.get('FACEBOOK_CLIENT_SECRET', '')
    id = "facebook"
    profile_url = GRAPH_API_URL + "/me"
    access_token_url = GRAPH_API_URL + "/oauth/access_token"
    public_key_url = "https://www.facebook.com/.well-known/oauth/openid/jwks/"
    access_token_method = "GET"

    def get_user_fields(self):
        """
        available fields: https://developers.facebook.com/docs/graph-api/reference/user/
        """
        fields = [
            "id",
            "email",
            "name",
            "first_name",
            "last_name",
            "picture"
        ]
        return fields

    def get_uid(self, data: dict[str:str]):
        access_token = data.get('access_token', None)
        if access_token:
            appsecret_proof = hmac.new(self.client_secret.encode("utf-8"),
                                       msg=access_token.encode("utf-8"),
                                       digestmod=hashlib.sha256).hexdigest()
            resp = requests.get(
                self.profile_url,
                params={
                    "fields": ",".join(self.get_user_fields()),
                    "access_token": access_token,
                    "appsecret_proof": appsecret_proof,
                },
            )
            resp.raise_for_status()
            data = resp.json()
            return data['id'], self.get_model_fields(data), data
        else:
            return super().get_uid(data)

    def get_model_fields(self, data):
        return dict(
            email=data.get("email"),
            last_name=data.get("last_name", ''),
            first_name=data.get("first_name", ''),
        )


import re

from django.contrib.auth import get_user_model
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.core import mail
from django.core.cache import cache
from django.db import transaction
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from oauth2_provider.models import get_access_token_model
from oauth2_provider.settings import oauth2_settings
from rest_framework import serializers

from auth_framework.serializers.mixin_serializers import EmailMixin
from auth_framework.serializers.password_serializers import PasswordChangeSerializer
from .test_oauth import BaseTest, UserInfoTestSerializer

AccessToken = get_access_token_model()
UserModel = get_user_model()


class EmailMixinTestSerializer(EmailMixin, UserInfoTestSerializer):
    pass


class BaseAccountTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.tok = AccessToken.objects.create(
            user=self.test_user,
            token="1234567890",
            application=self.application,
            expires=timezone.now() + timezone.timedelta(days=1),
            scope="read write",
        )

    def tearDown(self):
        super().tearDown()
        self.tok.delete()


class UserInfoTest(BaseAccountTest):

    @override_settings(AUTH_FRAMEWORK={"SERIALIZERS": {'USERINFO_SERIALIZER': EmailMixinTestSerializer}})
    def test_get_self_userinfo(self):
        self.client.credentials(HTTP_AUTHORIZATION='Bearer ' + self.tok.token)
        response = self.client.get(reverse("userinfo",args=['self']),)
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, {'first_name': '', 'last_name': '', 'email': 'test@example.com'})

    @override_settings(AUTH_FRAMEWORK={"UNIQUE_EMAIL": False,
                                    'SERIALIZERS': {'USERINFO_SERIALIZER': EmailMixinTestSerializer}})
    def test_patch_self_userinfo(self):
        self.client.credentials(HTTP_AUTHORIZATION='Bearer ' + self.tok.token)
        response = self.client.patch(reverse("userinfo", args=['self']),
                                     data={'last_name': 'patch',})
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, {'first_name': '', 'last_name': 'patch', 'email': 'test@example.com'})
        # since db migrations persist, we just check serializer validation
        serializer = EmailMixinTestSerializer(data={'email': 'dev@example.com',})
        self.assertEqual(serializer.is_valid(), True)

    @override_settings(AUTH_FRAMEWORK={"UNIQUE_EMAIL": True,
                                       'SERIALIZERS': {'USERINFO_SERIALIZER': EmailMixinTestSerializer}})
    def test_patch_unique_email(self):
        self.client.credentials(HTTP_AUTHORIZATION='Bearer ' + self.tok.token)
        response = self.client.patch(reverse("userinfo", args=['self']),
                                     data={'email': 'test@example.com'})
        self.assertDictEqual(response.data, {'first_name': '', 'last_name': '', 'email': 'test@example.com'})
        self.assertEqual(response.status_code, 200)
        response = self.client.patch(reverse("userinfo", args=['self']),
                                     data={'email': 'dev@example.com'})
        self.assertEqual(response.status_code, 400)

    @override_settings(AUTH_FRAMEWORK={"SERIALIZERS": {'USERINFO_SERIALIZER': EmailMixinTestSerializer}})
    def test_read_others_info(self):
        dev_pk=self.dev_user.pk
        self.client.credentials(HTTP_AUTHORIZATION='Bearer ' + self.tok.token)
        response = self.client.get(reverse("userinfo", args=[dev_pk]), )
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.data, {'first_name': '', 'last_name': '', 'email': 'dev@example.com'})

    def test_403_patch_others_info(self):
        dev_pk = self.dev_user.pk
        self.client.credentials(HTTP_AUTHORIZATION='Bearer ' + self.tok.token)
        response = self.client.patch(reverse("userinfo", args=[dev_pk]),
                                     data={'first_name': 'test', 'last_name': 'patch'})
        self.assertEqual(response.status_code, 403)

    def test_401_get_self_userinfo(self):
        response = self.client.get(reverse("userinfo",args=['self']),)
        self.assertEqual(response.status_code, 401)


class UserChangePasswordTest(BaseAccountTest):

    def _get_test_data(self):
        data = {
            'old_password': '123456',
            'password1': '1234qwes',
        }
        return data

    @override_settings(AUTH_FRAMEWORK={"USE_PASSWORD_TWICE_VALIDATION": False})
    def test_change_password(self):
        self.client.credentials(HTTP_AUTHORIZATION='Bearer ' + self.tok.token)
        response = self.client.post(reverse("change-password"), data=self._get_test_data())
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {"message": "New password has been saved."})

    @override_settings(AUTH_FRAMEWORK={"USE_PASSWORD_TWICE_VALIDATION": False})
    def test_400_change_password(self):
        self.client.credentials(HTTP_AUTHORIZATION='Bearer ' + self.tok.token)
        data = self._get_test_data()
        data |= {'old_password': 'NOT_MY_PASSWORD'}
        response = self.client.post(reverse("change-password"), data=data)
        self.assertEqual(response.status_code, 400)

    @override_settings(AUTH_FRAMEWORK={"USE_PASSWORD_TWICE_VALIDATION": True})
    def test_change_password_twice_validation(self):
        data = self._get_test_data()
        request = self.factory.post(reverse("change-password"))
        request.user = self.test_user
        serializer_kwargs = {
            'data': data,
            'context': {'request': request}
        }
        serializer = PasswordChangeSerializer(**serializer_kwargs)
        self.assertEqual(serializer.is_valid(), False)
        data |= {'password2':  '1234qwes',}
        serializer = PasswordChangeSerializer(**serializer_kwargs)
        self.assertEqual(serializer.is_valid(), True)
        data |= {'password2': '1234qwec', }
        serializer = PasswordChangeSerializer(**serializer_kwargs)
        self.assertEqual(serializer.is_valid(), False)

    def test_401_get_self_userinfo(self):
        response = self.client.post(reverse("change-password"))
        self.assertEqual(response.status_code, 401)


@override_settings(AUTH_FRAMEWORK={"USE_PASSWORD_TWICE_VALIDATION": False})
class UserResetPasswordTest(BaseAccountTest):
    """
        skip testing on send reset pin to phone
    """
    def test_pin_reset_entrypoint(self):
        self.assertEqual(reverse("reset-password-entrypoint", args=['pin']),
                         reverse("reset-password-entrypoint", kwargs={'method': 'pin'}))
        response = self.client.post(reverse("reset-password-entrypoint", kwargs={'method': 'pin'}),
                                    data={'email': self.test_user.email})
        self.assertEqual(response.status_code, 202)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, [self.test_user.email])
        pin_regex = re.search(r'reset pin: (?P<pin>\d+)',mail.outbox[0].body)
        self.assertIsNotNone(pin_regex)
        self.assertEqual(int(pin_regex.groupdict()['pin']), cache.get('pin_verify:%s' % self.test_user.email))

    def test_400_pin_reset(self):
        response = self.client.post(reverse("reset-password-entrypoint", kwargs={'method': 'pin'}),
                                    data={'email': 'NOT_existed@not.existed'})
        self.assertEqual(response.status_code, 400)

    def test_pin_reset_password(self):
        cache.set('pin_verify:%s' % self.test_user.email, 123456)
        response = self.client.post(reverse("pin-reset-password"),
                                    data={
                                        'email': self.test_user.email,
                                        'pin': 123456,
                                        'password1': 'resetofpin',})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {"message": "New password has been saved."})
        self.assertIsNone(cache.get('pin_verify:%s' % self.test_user.email))
        user = UserModel.objects.get(pk=self.test_user.pk)
        self.assertTrue(user.check_password('resetofpin'))

    def test_400_pin_reset_password(self):
        cache.set('pin_verify:%s' % self.test_user.email, 333333)
        response = self.client.post(reverse("pin-reset-password"),
                                    data={
                                        'email': self.test_user.email,
                                        'pin': 123456,
                                        'password1': 'resetofpin', })
        self.assertEqual(response.status_code, 400)
        response = self.client.post(reverse("pin-reset-password"),
                                    data={
                                        'email': self.dev_user.email,
                                        'pin': 333333,
                                        'password1': 'resetofpin', })
        self.assertEqual(response.status_code, 400)
        self.test_user.email = 'new@email.com'
        self.test_user.save()
        response = self.client.post(reverse("pin-reset-password"),
                                    data={
                                        'email': 'new@email.com',
                                        'pin': 333333,
                                        'password1': 'resetofpin', })
        self.assertEqual(response.status_code, 400)

    def test_link_reset_entrypoint(self):
        self.assertEqual(reverse("reset-password-entrypoint", args=['link']),
                         reverse("reset-password-entrypoint", kwargs={'method': 'link'}))

        response = self.client.post(reverse("reset-password-entrypoint", kwargs={'method': 'link'}),
                                    data={'email': self.test_user.email})
        self.assertEqual(response.status_code, 202)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, [self.test_user.email])
        body = mail.outbox[0].body
        self.assertGreater(body.find("/password/reset/"), 0)

        # Extract URL for `password_reset_from_key` view and access it
        url = body[body.find("/password/reset/"):].split()[0]
        response = self.client.get(url)
        self.assertEqual(response.data, {'message': 'valid url'})
        response = self.client.post(url, {'password1': 'resetoflink',})
        self.assertEqual(response.status_code, 200)
        user = UserModel.objects.get(pk=self.test_user.pk)
        self.assertTrue(user.check_password('resetoflink'))
        response = self.client.post(url, {'password1': 'resetoflink', })
        self.assertEqual(response.status_code, 400)


class CustomSignUpSerializer(serializers.Serializer):
    first_name = serializers.CharField(required=False)
    last_name = serializers.CharField()


class UsernameTestValidator(UnicodeUsernameValidator):
    regex = r'^\w+$'


@override_settings(AUTH_FRAMEWORK={"USE_PASSWORD_TWICE_VALIDATION": False,
                                   "SERIALIZERS": {'USERINFO_SERIALIZER': UserInfoTestSerializer}})
class CreateUserWithToken(BaseTest):

    def _hard_reload_serializer_mod(self):
        from importlib import reload
        from auth_framework.oauth import views
        from auth_framework.serializers import signup_serializers
        reload(signup_serializers)
        reload(views)

    def test_create_user(self):
        self._hard_reload_serializer_mod()
        url = reverse('create-user')
        data1 = {'email': 'adb@ad.ca', 'password1': 'qwed678s',}
        response = self.client.post(url, data1,)
        self.assertEqual(response.status_code, 201)
        self.assertIn('access_token', response.data)
        self.assertIn('refresh_token', response.data)
        self.assertEqual(response.data["expires_in"], oauth2_settings.ACCESS_TOKEN_EXPIRE_SECONDS)
        self.assertTrue(UserModel.objects.filter(email='adb@ad.ca').exists())
        self.assertDictEqual(response.data["user"], {'first_name': '', 'last_name': '', 'email': 'adb@ad.ca'})

    @override_settings(AUTH_FRAMEWORK={"USE_PASSWORD_TWICE_VALIDATION": False,
                                    "SERIALIZERS": {'USERINFO_SERIALIZER': UserInfoTestSerializer,
                                                    'SIGNUP_SERIALIZER': CustomSignUpSerializer
                                                    }})
    def test_400_custom_field_on_creation(self):
        self._hard_reload_serializer_mod()
        url = reverse('create-user')
        data = {
            'email': 'adb@ad.ca', 'password1': 'qwed678s',
            }
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 400)

    @override_settings(AUTH_FRAMEWORK={"SIGNUP_USERNAME_REQUIRED": True})
    def test_400_username_required_on_creation(self):
        url = reverse('create-user')
        data = {
            'email': 'adb@ad.ca', 'password1': 'qwed678s', 'password2': 'qwed678s',
            }
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 400)

    @override_settings(AUTH_FRAMEWORK={"SIGNUP_USERNAME_VALIDATORS":
                                        ['tests.test_account.UsernameTestValidator']})
    def test_custom_username_validator(self):
        url = reverse('create-user')
        data = {
            'email': 'adb@ad.ca', 'password1': 'qwed678s', 'password2': 'qwed678s', 'username':'invalid.dot',
        }
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 400)

    @override_settings(AUTH_FRAMEWORK={"USE_PASSWORD_TWICE_VALIDATION": False,
                                       "SERIALIZERS": {'USERINFO_SERIALIZER': UserInfoTestSerializer,
                                                       'SIGNUP_SERIALIZER': CustomSignUpSerializer
                                                       }})
    def test_custom_field_on_creation(self):
        self._hard_reload_serializer_mod()
        url = reverse('create-user')
        with transaction.atomic():
            data = {'email': 'adb@ad.ca', 'password1': 'qwed678s', 'last_name': 'custom'}
            response = self.client.post(url, data)
            self.assertEqual(response.status_code, 201)



__author__ = 'Carrycat'

import json
import time
import uuid

from django.db import transaction
from django.http import HttpRequest
from django.urls import NoReverseMatch
from django.utils import dateformat, timezone
from django.utils.functional import SimpleLazyObject
from jwcrypto import jwt
from jwcrypto.common import JWException
from jwcrypto.jwt import JWTExpired
from oauth2_provider.settings import oauth2_settings as oauth2_provider_settings, oauth2_settings
from oauth2_provider.exceptions import FatalClientError
from oauth2_provider.models import get_refresh_token_model
from oauth2_provider.oauth2_validators import OAuth2Validator as DefaultOAuth2Validator
from oauthlib.oauth2.rfc6749 import errors
from django.contrib.auth import authenticate

from ..backends.auth_backends import AuthenticationBackend
from ..settings import app_settings, import_callable

RefreshToken = get_refresh_token_model()


class LazyAuthenticatedUser(SimpleLazyObject):
    '''
                 ▄              ▄
                  ▌▒█           ▄▀▒▌
                  ▌▒▒█        ▄▀▒▒▒▐
                 ▐▄▀▒▒▀▀▀▀▄▄▄▀▒▒▒▒▒▐
               ▄▄▀▒░▒▒▒▒▒▒▒▒▒█▒▒▄█▒▐
             ▄▀▒▒▒░░░▒▒▒░░░▒▒▒▀██▀▒▌
            ▐▒▒▒▄▄▒▒▒▒░░░▒▒▒▒▒▒▒▀▄▒▒▌
            ▌░░▌█▀▒▒▒▒▒▄▀█▄▒▒▒▒▒▒▒█▒▐
           ▐░░░▒▒▒▒▒▒▒▒▌██▀▒▒░░░▒▒▒▀▄▌
           ▌░▒▄██▄▒▒▒▒▒▒▒▒▒░░░░░░▒▒▒▒▌
          ▌▒▀▐▄█▄█▌▄░▀▒▒░░░░░░░░░░▒▒▒▐
          ▐▒▒▐▀▐▀▒░▄▄▒▄▒▒▒▒▒▒░▒░▒░▒▒▒▒▌
          ▐▒▒▒▀▀▄▄▒▒▒▄▒▒▒▒▒▒▒▒░▒░▒░▒▒▐
           ▌▒▒▒▒▒▒▀▀▀▒▒▒▒▒▒░▒░▒░▒░▒▒▒▌
    '''
    is_authenticated = True
    is_active = True

    def __bool__(self):
        return True


class OauthValidator(DefaultOAuth2Validator):
    def validate_user(self, username, password, client, request, *args, **kwargs):
        error_msg = 'credentials not correct'
        email = getattr(request, 'email', None)
        phone_number = getattr(request, 'phone_number', None)
        credentials = {
            'email': email,
            'phone_number': phone_number,
            'username': username,
            'password': password
        }
        user = authenticate(**credentials)
        if not user:
            raise errors.InvalidGrantError(error_msg, request=request)
        request.user = user
        return True

    def validate_id_token(self, token, scopes, request):
        """
        When users try to access resources, check that provided id_token is valid
        """
        if not token:
            return False
        if app_settings.OAUTH_SAVE_ID_TOKEN:
            return super().validate_id_token(token, scopes, request)
        # validate application by client_id
        key = self._get_key_for_token(token)
        if not key:
            return False
        try:
            jwt_token = jwt.JWT(key=key, jwt=token)
            claims = json.loads(jwt_token.claims)
        except (JWException, JWTExpired):
            return False
        backend = AuthenticationBackend()
        request.user = LazyAuthenticatedUser(lambda: backend.get_user(claims['sub']))

        # this is needed by django rest framework
        request.access_token = token
        return True

    def _save_id_token(self, jti, request, expires, *args, **kwargs):
        if app_settings.OAUTH_SAVE_ID_TOKEN:
            super()._save_id_token(jti, request, expires, *args, **kwargs)

    def save_token(self, token, request, *args, **kwargs):
        super().save_token(token, request, *args, **kwargs)
        user_serializers = app_settings.SERIALIZERS.get('USERINFO_SERIALIZER')
        if user_serializers:
            token['user'] = user_serializers(instance=request.user).data
        if app_settings.USE_ID_TOKEN and oauth2_provider_settings.OIDC_ENABLED and request.client.algorithm:
            token['id_token'] = self.get_id_token(token, None, request, **kwargs)

    @transaction.atomic
    def save_jwt_token(self, token, request, *args, **kwargs):
        """
        Of course we dont save jwt code itself. we only update its refresh token
        """

        if "scope" not in token:
            raise FatalClientError("Failed to renew access token: missing scope")

        if request.grant_type == "client_credentials":
            request.user = None

        refresh_token_instance = getattr(request, "refresh_token_instance", None)

        refresh_token_code = token.get("refresh_token", None)
        if isinstance(refresh_token_instance, RefreshToken):
            if (
                    self.rotate_refresh_token(request)
                    and isinstance(refresh_token_instance, RefreshToken)
            ):
                refresh_token_instance = RefreshToken.objects.select_for_update().get(
                    id=refresh_token_instance.id
                )
                request.refresh_token_instance = refresh_token_instance
                try:
                    refresh_token_instance.revoke()
                except (RefreshToken.DoesNotExist):
                    pass
                else:
                    setattr(request, "refresh_token_instance", None)
        self._create_refresh_token(request, refresh_token_code, None)

    def get_id_token_dictionary(self, token, token_handler, request):
        """
        Get the claims to put in the ID Token.

        These claims are in addition to the claims automatically added by
        ``oauthlib`` - aud, iat, nonce, at_hash, c_hash.

        This function adds in iss, exp and auth_time, plus any claims added from
        calling ``get_oidc_claims()``
        """
        claims = self.get_oidc_claims(token, token_handler, request)

        expiration_time = timezone.now() + timezone.timedelta(seconds=oauth2_settings.ID_TOKEN_EXPIRE_SECONDS)
        # Required ID Token claims
        claims.update(
            **{
                "iss": self.get_oidc_issuer_endpoint(request),
                "exp": int(dateformat.format(expiration_time, "U")),
                "jti": str(uuid.uuid4()),
            }
        )

        return claims, expiration_time

    def get_id_token(self, token, token_handler, request, nonce=None):
        id_token = {}
        id_token['aud'] = request.client_id
        id_token['iat'] = int(time.time())
        if nonce is not None:
            id_token["nonce"] = nonce
        return self.finalize_id_token(id_token, token, token_handler, request)

    def get_oidc_issuer_endpoint(self, request):
        try:
            return super().get_oidc_issuer_endpoint(request)
        except (NoReverseMatch, TypeError):
            if hasattr(request, 'build_absolute_uri'):
                return request.build_absolute_uri('account/oauth')
            django_request = HttpRequest()
            django_request.META = request.headers
            django_request._scheme = 'https'
            return django_request.build_absolute_uri('account/oauth')

    def get_additional_claims(self, request):
        return {
            "email": request.user.email,
            "last_name": request.user.last_name,
            "first_name": request.user.first_name,
        }



import os

from ..abstract_adapter import OauthAdapter


class AppleOauthAdapter(OauthAdapter):
    client_id = os.environ.get('APPLE_CLIENT_ID', '')
    client_secret = os.environ.get('APPLE_CLIENT_SECRET', '')
    id = "apple"
    access_token_url = "https://appleid.apple.com/auth/token"
    authorize_url = "https://appleid.apple.com/auth/authorize"
    public_key_url = "https://appleid.apple.com/auth/keys"

    def get_model_fields(self, data:dict[str:str]) -> dict[str:str]:
        return dict(
            email=data.get("email"),
            first_name=data.get('name', {}).get('first_name', ''),
            last_name=data.get('name', {}).get('last_name', '')
        )


import json

from django.contrib.auth import logout
from oauth2_provider.models import get_application_model
from oauth2_provider.views.mixins import OAuthLibMixin
from rest_framework import status
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.views import APIView

from .utils import get_access_token
from ..serializers.signup_serializers import SignUpSerializer
from ..social.providers.abstract_adapter import deserialize_social_acct
from ..social.serializers import SocialSignUpSerializer

Application = get_application_model()


class RevokeTokenView(APIView, OAuthLibMixin):
    # permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        url, headers, body, code_status = self.create_revocation_response(request)
        content = {}
        if body:
            content = json.loads(body)
            if 'error_description' in content:
                content = {'message': content['error_description']}
        try:
            logout(request)
        except AttributeError:
            # no session object in Request, pypass
            pass
        return Response(content, status=code_status)


class GrantUserWToken(APIView, OAuthLibMixin):

    def post(self, request, *args, **kwargs):
        if not request.data.get('username'):
            if hasattr(request.data, '_mutable'):
                request.data._mutable = True
            request.data['username'] = 'dummy'  # NOQA: ResourceOwnerPasswordCredentialsGrant requires username
        url, headers, body, code_status = self.create_token_response(request)
        content = json.loads(body)
        if code_status == 200:
            return Response(content)
        return Response({'message': content.get('error_description', content['error'])}, status=code_status)


class CreateUserWToken(GenericAPIView):
    # required_scopes = ['account.signup']

    def get_serializer_class(self):
        if 'social_login' in self.request.data:  # originally was .POST
            return SocialSignUpSerializer
        else:
            return SignUpSerializer

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # TODO: reedit
        if 'social_login' in self.request.data:  # originally was .POST but not friendly for Web Api
            context.update({'social_data': deserialize_social_acct(json.loads(self.request.data['social_login']))})
        return context

    def validate_client(self, raise_exception=False):
        # TODO: less confidential skip validate client(validate_client_id) for performance
        _errors = ""
        client_id = self.request.data.get('client_id')
        if not client_id:
            client = Application.objects.get(pk=1)
        else:
            client = Application.objects.get(client_id=client_id)
        if _errors and raise_exception:
            raise AuthenticationFailed(_errors)
        self.request.client = client
        self.request.client_id = client.client_id
        return client

    def post(self, *args, **kwargs):

        client = self.validate_client(raise_exception=True)
        serializer = self.get_serializer(data=self.request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.save()
        token = get_access_token(user, request=self.request)
        # token_serializer = AccessTokenSerializer(token)
        return Response(token, status=status.HTTP_201_CREATED)


from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.utils.http import base36_to_int
from rest_framework import status
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.response import Response

from ..serializers.password_serializers import PasswordChangeSerializer, CreateResetPinSerializer, \
    CreateResetLinkSerializer, ResetPasswordByPinSerializer, ResetPasswordByLinkSerializer, VerifyPinSerializer
from ..settings import app_settings

User = get_user_model()


class ChangePassword(GenericAPIView):
    """
    Accepts the following POST parameters: new_password1, new_password2
    Returns the success/fail message.
    """

    permission_classes = (IsAuthenticated,)
    serializer_class = PasswordChangeSerializer

    def post(self, request):
        serializer = self.get_serializer(data=request.data, context={'request': request})
        if serializer.is_valid(raise_exception=True):
            serializer.save()
            return Response({"message": "New password has been saved."})


class CreateResetEntryPoint(GenericAPIView):
    """
    Create a temporary pin for password reset.
    pin will be sent on save() by sms/email which configured on settings.py(WITH_PHONENUMBER_FIELD)
    Returns the success/fail message.
    """

    def get_serializer_class(self):
        try:
            method = self.kwargs.get('method') or self.args[0]
        except IndexError:
            method = app_settings.RESET_PASSWORD_BY
        if method == 'pin':
            return CreateResetPinSerializer
        else:
            return CreateResetLinkSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, check_existed=True)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({"message": 'Message has been sent'}, status=status.HTTP_202_ACCEPTED)


class CreateSignupPinSerializer(GenericAPIView):
    serializer_class = CreateResetPinSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, check_existed=False)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({"message": 'Message has been sent'}, status=status.HTTP_202_ACCEPTED)


class ResetPasswordByPin(GenericAPIView):
    permission_classes = (AllowAny,)
    serializer_class = ResetPasswordByPinSerializer

    def post(self, request, *args, **kwargs):
        is_verified = request.GET.get('is_verified', False)
        # Create a serializer with request.data
        serializer = self.get_serializer(data=request.data, skip_pin_validation=is_verified)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({"message": "New password has been saved."})


class ResetPasswordByLink(GenericAPIView):
    serializer_class = ResetPasswordByLinkSerializer
    permission_classes = (AllowAny,)

    def get_reset_user(self, uidb36):
        try:
            pk = base36_to_int(uidb36)
            return User.objects.get(pk=pk)
        except (ValueError, User.DoesNotExist):
            return None

    def is_link_valid(self, uidb36, token):
        user = self.get_reset_user(uidb36)
        if user is None or not default_token_generator.check_token(user, token):
            return Response({'message': "The password reset token was invalid."},
                            status=status.HTTP_400_BAD_REQUEST)
        return user

    def get(self, request, uidb36, token):
        resp = self.is_link_valid(uidb36, token)
        if isinstance(resp,Response):
            return resp
        else:
            return Response({'message': 'valid url'})

    def post(self, request, *args, **kwargs):
        resp = self.is_link_valid(kwargs.get('uidb36'), kwargs.get('token'))
        if isinstance(resp, Response):
            return resp
        reset_user = resp
        context = self.get_serializer_context()
        context.update({'user': reset_user})
        serializer = self.get_serializer(data=request.data, context=context)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({"message": "New password has been saved."})


class VerifyPin(GenericAPIView):

    def post(self, request, *args, **kwargs):
        phone_unique = request.GET.get('phone_unique', False)
        keep_alive = request.GET.get('keep_alive', False)
        context = self.get_serializer_context()
        serializer = VerifyPinSerializer(data=request.data,
                                         keep_alive=keep_alive,
                                         phone_unique=phone_unique,
                                         context=context)
        if serializer.is_valid(raise_exception=True):
            return Response({"result": "valid"})



__author__ = 'Carrycat'
from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.generics import RetrieveUpdateAPIView, get_object_or_404, GenericAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from ..permissions import IsSelfOrReadOnly
from ..serializers.account_serializers import UpdatePhoneSerializer, DefaultUserinfoSerializer
from ..settings import app_settings

User = get_user_model()


class UserDetails(RetrieveUpdateAPIView):
    permission_classes = (IsAuthenticated, IsSelfOrReadOnly)  # for scopes favor: ScopesIsSelfOrReadOnly,
    queryset = User.objects.filter(is_active=True)
    # required_scopes = ['account.get_user_details']

    def get_serializer_class(self):
        return app_settings.SERIALIZERS.get('USERINFO_SERIALIZER') or DefaultUserinfoSerializer

    def get_object(self):
        if self.kwargs['pk'] == 'self':
            obj = self.request.user
        else:
            obj = get_object_or_404(User, pk=self.kwargs['pk'])
        self.check_object_permissions(self.request, obj)
        return obj

    def get(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        data = serializer.data
        return Response(data)


class EmailCheck(GenericAPIView):
    def post(self, *args, **kwargs):
        exist = User.objects.filter(email=self.request.data.get('email', '')).exists()
        return Response(not exist)


class PhoneChangeView(GenericAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = UpdatePhoneSerializer

    def post(self, request):
        serializer = self.get_serializer(data=request.data, context={'request': request})
        if serializer.is_valid(raise_exception=True):
            serializer.save()
            return Response({"message": "New phone number has been saved."})


__author__ = 'Carrycat'

from django.contrib.auth import authenticate, get_user_model
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import AuthenticationFailed

from ..models import settings_default_value
from .mixin_serializers import PhoneNumPinMixin, EmailMixin, PhoneNumberField

User = get_user_model()


class UpdatePhoneSerializer(PhoneNumPinMixin, serializers.Serializer):
    pin = serializers.IntegerField(max_value=999999)

    def __init__(self, *args, **kwargs):
        kwargs.update({'check_existed': False})
        request = self.context.get('request')
        self.user = getattr(request, 'user', None)
        super().__init__(**kwargs)

    def save(self):
        self.user.phone_number = self.validated_data.get("phone_number")
        self.user.save()
        return self.user


class SessionLoginSerializer(serializers.Serializer):
    username = serializers.CharField(required=False)
    email = serializers.CharField(required=False)
    phone_number = PhoneNumberField(required=False)
    password = serializers.CharField()

    def validate(self, attrs):
        request = self.context.get('request')
        user = authenticate(request=request, **attrs)
        if not user:
            raise AuthenticationFailed('Unable to log in with provided credentials.')
        return user


class DefaultUserinfoSerializer(EmailMixin, serializers.ModelSerializer):
    # avatar = serializers.URLField(source='avatar_url', read_only=True)  # api_avatar_url
    phone_number = PhoneNumberField(read_only=True)

    class Meta:
        model = User
        fields = User.get_fields()

    def validate_username(self, value):
        user = self.context['request'].user
        if not User.objects.exclude(pk=user.pk).filter(**{'username': value}).exists():
            return value
        raise serializers.ValidationError(_("A user is already registered with this username."), code='AA001')

    def update(self, instance, validated_data):
        if 'avatar' in validated_data:
            # mark avatar as a update_field, since the kwarg 'update_fields'
            # cannot be passed to instance.save() from DRF.
            instance._avatar_file_updated = True
        return super().update(instance, validated_data)


__author__ = 'Carrycat'
from oauth2_provider.oauth2_backends import OAuthLibCore as DefaultOAuthLibCore


class OAuthLibCore(DefaultOAuthLibCore):
    """
    Extends the default OAuthLibCore to parse correctly requests with application/json Content-Type
    """
    def extract_body(self, request):
        """
        Using data instead of DATA, to support accept JSON body
        """
        if isinstance(request.data, dict):
            body = request.data.items()
        else:
            body = {'data': request.data}.items()
        return body



import pytest
from django.contrib.auth import get_user_model
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from oauth2_provider.models import get_application_model, get_access_token_model, get_refresh_token_model
from oauth2_provider.settings import oauth2_settings
from rest_framework import serializers
from rest_framework.test import APITestCase, APIRequestFactory


Application = get_application_model()
AccessToken = get_access_token_model()
RefreshToken = get_refresh_token_model()
UserModel = get_user_model()


class UserInfoTestSerializer(serializers.ModelSerializer):

    class Meta:
        model = UserModel
        fields = ('first_name', 'last_name', 'email',)


@pytest.mark.django_db
class BaseTest(APITestCase):
    def setUp(self):
        self.factory = APIRequestFactory()
        self.test_user = UserModel.objects.create_user("test_user", "test@example.com", "123456",)
        self.dev_user = UserModel.objects.create_user("dev_user", "dev@example.com", "123456",)
        self.client_secret = "1234567890abcdefghijklmnopqrstuvwxyz"
        self.application = Application.objects.create(
            name="Test Password Application",
            user=self.dev_user,
            client_type=Application.CLIENT_CONFIDENTIAL,
            client_secret=self.client_secret,
            authorization_grant_type=Application.GRANT_PASSWORD,
        )

    def tearDown(self):
        self.application.delete()
        self.test_user.delete()
        self.dev_user.delete()


class AccessTokenTest(BaseTest):

    def test_401_invalid_client(self):
        """
        Request an access token using Resource Owner Password Flow
        """
        token_request_data = {
            "grant_type": "password",
            "username": "test_user",
            "password": "123456",
            "client_id": self.application.client_id,
        }

        response = self.client.post(reverse("token"), data=token_request_data)
        self.assertEqual(response.status_code, 401)
        token_request_data |= {
            "client_secret": self.client_secret,
            "client_id": "wrong_client_id"
        }
        response = self.client.post(reverse("token"), data=token_request_data)
        self.assertEqual(response.status_code, 401)

    @override_settings(AUTH_FRAMEWORK={"SERIALIZERS": {'USERINFO_SERIALIZER': None}})
    def test_get_token_only(self):
        """
        Request an access token using Resource Owner Password Flow
        """
        token_request_data = {
            "grant_type": "password",
            "username": "test_user",
            "password": "123456",
            "client_id": self.application.client_id,
            "client_secret": self.client_secret
        }

        response = self.client.post(reverse("token"), data=token_request_data)
        self.assertEqual(response.status_code, 200)
        content = response.data
        self.assertEqual(content["token_type"], "Bearer")
        self.assertEqual(content["expires_in"], oauth2_settings.ACCESS_TOKEN_EXPIRE_SECONDS)
        self.assertIn('access_token', content)
        self.assertIn('refresh_token', content)
        self.assertNotIn('user', content)

    @override_settings(AUTH_FRAMEWORK={"SERIALIZERS": {'USERINFO_SERIALIZER': UserInfoTestSerializer}})
    def test_get_token_with_userinfo(self):
        """
        Request an access token using Resource Owner Password Flow
        """
        token_request_data = {
            "grant_type": "password",
            "username": "test_user",
            "password": "123456",
            "client_id": self.application.client_id,
            "client_secret": self.client_secret
        }

        response = self.client.post(reverse("token"), data=token_request_data)
        self.assertEqual(response.status_code, 200)
        content = response.data
        self.assertIn('access_token', content)
        self.assertIn('refresh_token', content)
        self.assertDictEqual(content["user"], {'first_name': '', 'last_name': '', 'email': 'test@example.com'})

    def test_get_token_by_email(self):
        """
        Request an access token using Resource Owner Password Flow
        """
        token_request_data = {
            "grant_type": "password",
            "email": "test@example.com",
            "password": "123456",
            "client_id": self.application.client_id,
            "client_secret": self.client_secret
        }

        response = self.client.post(reverse("token"), data=token_request_data)
        self.assertEqual(response.status_code, 200)
        self.assertIn('access_token', response.data)

    def test_400_credentials(self):
        """
        Request an access token using Resource Owner Password Flow
        """
        token_request_data = {
            "grant_type": "password",
            "username": "test_user",
            "password": "NOT_MY_PASS",
            "client_id": self.application.client_id,
            "client_secret": self.client_secret
        }

        response = self.client.post(reverse("token"), data=token_request_data)
        self.assertEqual(response.status_code, 400)

    def test_revoke_access_token(self):
        tok = AccessToken.objects.create(
            user=self.test_user,
            token="1234567890",
            application=self.application,
            expires=timezone.now() + timezone.timedelta(days=1),
            scope="read write",
        )

        data = {
            "client_id": self.application.client_id,
            "client_secret": self.client_secret,
            "token": tok.token,
        }
        url = reverse("revoke-token")
        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {})
        self.assertFalse(AccessToken.objects.filter(id=tok.id).exists())

    def test_revoke_refresh_token(self):
        tok = AccessToken.objects.create(
            user=self.test_user,
            token="1234567890",
            application=self.application,
            expires=timezone.now() + timezone.timedelta(days=1),
            scope="read write",
        )
        rtok = RefreshToken.objects.create(
            user=self.test_user, token="999999999", application=self.application, access_token=tok
        )

        data = {
            "client_id": self.application.client_id,
            "client_secret": self.client_secret,
            "token": rtok.token,
        }

        url = reverse("revoke-token")
        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, 200)
        refresh_token = RefreshToken.objects.filter(id=rtok.id).first()
        self.assertIsNotNone(refresh_token.revoked)
        self.assertFalse(AccessToken.objects.filter(id=rtok.access_token.id).exists())

    def test_refresh_token_grant(self):
        tok = AccessToken.objects.create(
            user=self.test_user,
            token="1234567890",
            application=self.application,
            expires=timezone.now() + timezone.timedelta(days=1),
            scope="read write",
        )
        rtok = RefreshToken.objects.create(
            user=self.test_user, token="999999999", application=self.application, access_token=tok
        )
        token_request_data = {
            "grant_type": "refresh_token",
            "client_id": self.application.client_id,
            "client_secret": self.client_secret,
            'refresh_token': rtok.token
        }
        response = self.client.post(reverse("token"), data=token_request_data)
        self.assertEqual(response.status_code, 200)
        content = response.data
        self.assertEqual(content["token_type"], "Bearer")
        self.assertEqual(content["expires_in"], oauth2_settings.ACCESS_TOKEN_EXPIRE_SECONDS)
        self.assertNotEqual(content['access_token'], tok.token)
        self.assertNotEqual(content['refresh_token'], rtok.token)
        refresh_token = RefreshToken.objects.filter(id=rtok.id).first()
        self.assertIsNotNone(refresh_token.revoked)
        self.assertFalse(AccessToken.objects.filter(id=rtok.access_token.id).exists())




from django.contrib.auth.backends import ModelBackend as DjangoModelBackend
from django.contrib.auth import get_user_model
try:
    from phonenumber_field.phonenumber import to_python
except ImportError:
    def to_python(data):
        return str(data)

User = get_user_model()


class AuthenticationBackend(DjangoModelBackend):

    def authenticate(self, request, password=None, **kwargs):
        username = kwargs.get("username")
        email = kwargs.get("email")
        phone_number = to_python(kwargs['phone_number']) if kwargs.get("phone_number") else None

        try:
            if phone_number and phone_number.is_valid():
                user = User.objects.get(phone_number=phone_number)
            elif email:
                user = User.objects.get(email__iexact=email)
            elif username:
                user = User.objects.get(username=username)
            else:
                return None
        except User.DoesNotExist:
            User().set_password(password)
        else:
            if user.check_password(password) and self.user_can_authenticate(user):
                return user



__author__ = 'Carrycat'
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from oauth2_provider.models import AccessToken

from ..settings import app_settings, import_callable

# user_serializers = import_callable(app_settings.SERIALIZERS.get('USERINFO_SERIALIZER'))


class AccessTokenSerializer(serializers.ModelSerializer):
    access_token = serializers.CharField(source='token')
    refresh_token = serializers.CharField(source='refresh_token.token')
    expires_in = serializers.DateTimeField(source='expires')

    def __init__(self, *args, **kwargs):
        if app_settings.SERIALIZERS.get('USERINFO_SERIALIZER'):
            self.fields['user'] = app_settings.SERIALIZERS.get('USERINFO_SERIALIZER')()
        super().__init__(*args, **kwargs)

    class Meta:
        model = AccessToken
        fields = ('user', 'access_token', 'refresh_token', 'expires_in', 'scope')
        read_only_fields = ('user', 'access_token', 'refresh_token', 'expires_in', 'scope')



__author__ = 'Carrycat'

from django.urls import re_path
from oauth2_provider.views import JwksInfoView

from . import views

urlpatterns = [
    re_path(r'^token/$', views.GrantUserWToken.as_view(), name="token"),
    re_path(r'^revoke/$', views.RevokeTokenView.as_view(), name="revoke-token"),
    # reuse oauth2_provider jwks url
    re_path(r"^\.well-known/jwks.json$", JwksInfoView.as_view(), name="jwks-info"),
]



from django.urls import re_path, include

from .oauth.views import CreateUserWToken
from .settings import app_settings
from .views import password_views, account_views, session_views

urlpatterns = [
    re_path('login/', session_views.SessionLogin.as_view(), name="account_login"),
    re_path('logout/', session_views.SessionLogout.as_view(), name="account_logout"),
    re_path('password/change/', password_views.ChangePassword.as_view(), name='change-password'),
    re_path('password/reset/pin/$', password_views.ResetPasswordByPin.as_view(), name='pin-reset-password'),
    re_path(r'^password/(?P<method>link|pin)/$', password_views.CreateResetEntryPoint.as_view(),
            name='reset-password-entrypoint'),
    re_path(r"^password/reset/(?P<uidb36>[0-9A-Za-z]+)-(?P<token>.+)/$",
            password_views.ResetPasswordByLink.as_view(),
            name="link-reset-password"),
    re_path(r'^registration/$', CreateUserWToken.as_view(), name="create-user"),
    re_path('registration/pin/', password_views.CreateSignupPinSerializer.as_view(), name="create-signup-pin"),
    re_path('email-verify/', account_views.EmailCheck.as_view()),
    re_path(r'^(?P<pk>\d+|self)/$', account_views.UserDetails.as_view(), name="userinfo"),
    re_path('oauth/', include('auth_framework.oauth.urls')),
    re_path('social/', include('auth_framework.social.urls')),
    re_path('pin/verify/', password_views.VerifyPin.as_view(), name="verify_pin_api"),
]
if app_settings.USE_PHONENUMBER_FIELD:
    urlpatterns += [
        re_path('phone/', account_views.PhoneChangeView.as_view(), name="edit_phone"),
    ]



from io import BytesIO
from django.conf import settings
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.core.files import File
from django.db import models
from django.contrib.auth.models import AbstractUser as DjangoAbstractUser
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _
try:
    from PIL import Image
except ImportError:
    Image = None
from .settings import app_settings


class CustomUnicodeUsernameValidator(UnicodeUsernameValidator):
    regex = r'^[\w.-]+$'
    message = _(
        'Enter a valid username. This value may contain only letters, '
        'numbers, and ./-/_ characters.'
    )


class DefaultAbstractUser(DjangoAbstractUser):
    username = models.CharField(_('username'), max_length=30, unique=True, db_collation='ci',
                                help_text=_('Required. 30 characters or fewer. Letters, digits and /./-/_ only.'),
                                validators=[CustomUnicodeUsernameValidator],
                                error_messages={'unique': _("A user with that username already exists."),},
                                )
    if app_settings.UNIQUE_EMAIL:
        email = models.EmailField(_('email address'), unique=True, db_collation='ci',)

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')
        abstract = True


def avatar_dir_path(instance, filename):
    # add suffix to avoid unnecessary cache on CDN
    return 'avatars/{pk}_{suffix}.jpeg'.format(pk=instance.pk, suffix=get_random_string(5))


class AvatarMixin(models.Model):
    avatar = models.ImageField(_("logo"), upload_to=avatar_dir_path, blank=True, max_length=300,)

    @property
    def avatar_url(self):
        if self.avatar:
            return self.avatar.url
        else:
            return None

    def save(self,  *args, **kwargs):
        if Image and self.avatar and getattr(self, '_avatar_file_updated', False):
            avatar = Image.open(self.avatar)
            if avatar.mode != 'RGB':
                # Scale below in other worker or Taskqueue(Celery), when needed
                jpeg_image = avatar.convert('RGB')
                image_io = BytesIO()
                jpeg_image.save(image_io, format='JPEG', quality=90)
                self.avatar = File(image_io, name='avatar.jpeg')
        return super().save(*args, **kwargs)

    class Meta:
        abstract = True


try:
    from phonenumber_field.modelfields import PhoneNumberField

    class PhoneNumberMixin(models.Model):
        phone_number = PhoneNumberField(unique=True)

        class Meta:
            abstract = True
except ImportError:
    class PhoneNumberMixin(object):
        pass

class AbstractSocialAccount(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='socialaccount_set', on_delete=models.CASCADE)
    provider = models.CharField(_("provider slug"),max_length=30,)
    uid = models.CharField(_("uid"), max_length=191)
    date_joined = models.DateTimeField(_("date joined"), auto_now_add=True)
    extra_data = models.JSONField(verbose_name=_("extra data"), default=dict)

    class Meta:
        unique_together = ("provider", "uid")
        verbose_name = _("social account")
        verbose_name_plural = _("social accounts")
        abstract = True

    def __str__(self):
        return "{user} - {provider}".format(user=self.user, provider=self.provider)



from oauth2_provider.contrib.rest_framework import TokenHasScope
from rest_framework import permissions


import logging

log = logging.getLogger('oauth2_provider')


class IsAuthenticatedAndActive(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.user and request.user.is_authenticated and request.user.is_active


class IsAdminOrIsSelf(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj.user == request.user or request.user.is_staff or request.user.is_superuser


class IsSelfOrReadOnly(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        if request.method in permissions.SAFE_METHODS:
            return True
        else:
            return obj == request.user


class ScopesIsSelfOrReadOnly(TokenHasScope):

    def has_permission(self, request, view):
        token = request.auth

        if not token:
            return False

        if hasattr(token, 'scope'):  # OAuth 2
            required_scopes = self.get_scopes(request, view)
            log.debug("Required scopes to access resource: {0}".format(required_scopes))

            return token.is_valid(required_scopes)

        assert False, ('TokenHasScope requires either the'
                       '`oauth2_provider.rest_framework.OAuth2Authentication` authentication '
                       'class to be used.')

    def has_object_permission(self, request, view, obj):
        if request.method in permissions.SAFE_METHODS:
            return self.has_permission(request, view)
        else:
            return obj == request.user


# maybe only using for testing...
class IsAdminOrScopesSelfOrReadOnly(ScopesIsSelfOrReadOnly):

    def has_permission(self, request, view):
        token = request.auth

        if request.user and request.user.is_authenticated and request.user.is_superuser:
            return True
        elif not token:
            return False

        if hasattr(token, 'scope'):  # OAuth 2
            required_scopes = self.get_scopes(request, view)
            log.debug("Required scopes to access resource: {0}".format(required_scopes))

            return token.is_valid(required_scopes)

        assert False, ('TokenHasScope requires either the'
                       '`oauth2_provider.rest_framework.OAuth2Authentication` authentication '
                       'class to be used.')



from random import randint

from django.urls import reverse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.shortcuts import get_current_site
from django.core.cache import cache
from django.utils.http import int_to_base36
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from ..exceptions import AuthFrameworkImproperlyConfigured
from .mixin_serializers import PasswordMixin, EmailMixin, PhoneNumMixin, PhoneNumPinMixin
from ..settings import app_settings
from ..utils import render_mail

User = get_user_model()


def send_email(email, **extra_kwargs) -> None:
    if extra_kwargs.get('template', None):
        if app_settings.USE_CELERY_EMAIL:
            try:
                from auth_framework.tasks import send_email_task
                send_email_task.apply_async([email, ], kwargs=extra_kwargs)
                return
            except ImportError:
                pass
        template = extra_kwargs.pop('template', None)
        subject = extra_kwargs.pop('subject', 'No Title')
        msg = render_mail(subject, email, extra_kwargs, template)
        msg.send()


class PasswordChangeSerializer(PasswordMixin, serializers.Serializer):
    old_password = serializers.CharField()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = getattr(self.context.get('request'), 'user', None)

    def validate_old_password(self, old_password):
        invalid_password_conditions = (
            self.user,
            not self.user.check_password(old_password)
        )

        if all(invalid_password_conditions):
            raise serializers.ValidationError({"message": _('Invalid password'), "code": "AP003"})
        return old_password

    def save(self):
        password = self.validated_data.get('password')
        self.user.set_password(password)
        self.user.save()


class CreateResetPinSerializer(PhoneNumMixin, EmailMixin, serializers.Serializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if app_settings.USE_PHONENUMBER_FIELD:
            del self.fields['email']
            self.fields['phone_number'].required = True
            self.id_field = 'phone_number'
        else:
            if not app_settings.UNIQUE_EMAIL:
                raise AuthFrameworkImproperlyConfigured('Failed To determine the identity')
            del self.fields['phone_number']
            self.fields['email'].required = True
            self.id_field = 'email'

    def get_email_templ_kwargs(self, pin: int, email_to: str) -> dict:
        current_site = get_current_site(self.context['request'])
        return {
            'current_site': {'name': app_settings.TITLE_IN_MESSAGE or current_site.name, 'domain': current_site.domain},
            'template': 'auth/email_pin.txt',
            'subject': 'Set Up Your New Password',
            'pin': pin,
            # ...other_kwargs but 'email'
        }

    def send_reset_pin_email(self, pin, email):
        extra_kwargs = self.get_email_templ_kwargs(pin, email)
        send_email(email, **extra_kwargs)

    def save(self, **kwargs):
        pin = randint(10 ** 5, 10 ** 6 - 1)
        if self.id_field == 'phone_number':
            try:
                import os
                from twilio.base.exceptions import TwilioException, TwilioRestException
                from twilio.rest import Client
                account_sid = os.environ.get('TWILIO_CLIENT_ID')
                auth_token = os.environ.get('TWILIO_CLIENT_SECRET')
                from_number = os.environ.get('TWILIO_FROM_NUMBER')
                client = None
                if account_sid and auth_token:
                    client = Client(account_sid, auth_token)

                message = client.messages.create(
                    body="{title}: Your verification code is {pin}".format(
                        title=app_settings.TITLE_IN_MESSAGE or get_current_site(self.context['request']),
                        pin=pin),
                    from_=from_number,
                    to=str(self.validated_data['phone_number'])
                )
                cache.set('pin_verify:%s' % self.validated_data['phone_number'], pin, 60 * 10)
            except ImportError as err:
                raise AuthFrameworkImproperlyConfigured('Missing Twilio package for sending sms for pin: %s' % err)
            except AttributeError as err:
                raise AuthFrameworkImproperlyConfigured(
                    'Missing Twilio Env Variables for sending sms for pin: %s' % err)
            except TwilioRestException:
                raise serializers.ValidationError(
                    _("%s, is not a correct mobile number." % str(self.validated_data['phone_number'])), code='AP004')
        else:  # we choose classic email to send ping
            self.send_reset_pin_email(pin, self.validated_data['email'])
            cache.set('pin_verify:%s' % self.validated_data['email'], pin, 60 * 10)


class ResetPasswordByPinSerializer(PhoneNumMixin, EmailMixin, PasswordMixin, serializers.Serializer):
    email = serializers.EmailField(required=False)
    pin = serializers.IntegerField(max_value=999999)

    def __init__(self, **kwargs):
        kwargs.update({'check_existed': True})
        self.skip_pin_validation = getattr(self, 'skip_pin_validation', False) or \
                                   kwargs.pop('skip_pin_validation', False)
        self.user_found = None
        super().__init__(**kwargs)
        if app_settings.USE_PHONENUMBER_FIELD:
            del self.fields['email']
            self.fields['phone_number'].required = True
            self.id_field = 'phone_number'
        else:
            if not app_settings.UNIQUE_EMAIL:
                raise ProcessLookupError('Failed To determine the identity')
            del self.fields['phone_number']
            self.fields['email'].required = True
            self.id_field = 'email'
        self.user = None

    def validate(self, attrs):
        attrs = super().validate(attrs)
        pin = attrs.get("pin")
        lookup_field = attrs.get(self.id_field)
        if self.skip_pin_validation:
            if not cache.get('is_verified:%s' % lookup_field):
                raise serializers.ValidationError({'pin': {
                    'message': _("Session expired, please get a new pin"), 'code': 'AP006'}})
            cache.delete('is_verified:%s' % lookup_field)
        elif cache.get('pin_verify:%s' % lookup_field) == pin or \
                app_settings.DEBUG_PIN and pin == app_settings.DEBUG_PIN:
            cache.delete('pin_verify:%s' % lookup_field)
        else:
            raise serializers.ValidationError({'pin': {'message': _('pins do not match, please try to get new pin.'),
                                                       'code': 'AP05'}})
        return attrs

    def save(self):
        password = self.validated_data.get('password')
        self.user_found.set_password(password)
        self.user_found.save()


class CreateResetLinkSerializer(EmailMixin, serializers.Serializer):
    """
    Serializer for requesting a password reset e-mail.
    """

    def __init__(self, **kwargs):
        kwargs.update({'check_existed': True})
        self.user_found = None
        super().__init__(**kwargs)

    def get_email_templ_kwargs(self, uidb36: str, token: str, email_to: str, user: User) -> dict:
        request = self.context.get('request')
        current_site = get_current_site(request)
        path = reverse("link-reset-password", kwargs={'uidb36': uidb36, 'token': token})
        url = '{proto}://{domain}{url}'.format(
            proto=request.scheme,
            domain=current_site.domain,
            url=path)

        return {
            'current_site': {'name': current_site.name, 'domain': current_site.domain},
            "password_reset_url": url,
            'template': 'auth/email_reset_link.txt',
            'subject': 'Set Up Your New Password',
            'user': {
                'username': user.username,
                'full_name': user.get_full_name()
            },
            # ...other_kwargs
        }

    def save(self):
        user = self.user_found
        token = default_token_generator.make_token(user)
        uidb36 = int_to_base36(user.pk)
        email = self.validated_data["email"]
        extra_kwargs = self.get_email_templ_kwargs(uidb36, token, email, user)
        send_email(email, **extra_kwargs)
        return email


class ResetPasswordByLinkSerializer(PasswordMixin, serializers.Serializer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = self.context.get('user')

    def save(self):
        password = self.validated_data.get('password')
        self.user.set_password(password)
        self.user.save()


class VerifyPinSerializer(PhoneNumPinMixin, serializers.Serializer):
    pin = serializers.IntegerField(max_value=999999)

    def __init__(self, *args, **kwargs):
        self.phone_unique = kwargs.pop('phone_unique', False)
        self.keep_alive = kwargs.pop('keep_alive', False)
        self.skip_phone_validation = True
        super().__init__(**kwargs)
        # request = self.context.get('request')



import json

from django.contrib.auth import get_user_model
from django.test import override_settings
from django.urls import reverse
from rest_framework import serializers

from auth_framework.models import get_socialaccount_model
from auth_framework.social.providers.facebook.adapter import GRAPH_API_URL
from .test_oauth import BaseTest
from unittest.mock import patch, Mock

UserModel = get_user_model()
SocialAccountModel = get_socialaccount_model()

TEST_KEY_PUB = '''-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnzyis1ZjfNB0bBgKFMSv
vkTtwlvBsaJq7S5wA+kzeVOVpVWwkWdVha4s38XM/pa/yr47av7+z3VTmvDRyAHc
aT92whREFpLv9cj5lTeJSibyr/Mrm/YtjCZVWgaOYIhwrXwKLqPr/11inWsAkfIy
tvHWTxZYEcXLgAXFuUuaS3uF9gEiNQwzGTU1v0FqkqTBr4B8nW3HCN47XUu0t8Y0
e+lf4s4OxQawWD79J9/5d3Ry0vbV3Am1FtGJiJvOwRsIfVChDpYStTcHTCMqtvWb
V6L11BWkpzGXSW4Hv43qa+GSYOD2QU68Mb59oSk2OB+BtOLpJofmbGEGgvmwyCI9
MwIDAQAB
-----END PUBLIC KEY-----'''

TEST_KEY = '''-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAnzyis1ZjfNB0bBgKFMSvvkTtwlvBsaJq7S5wA+kzeVOVpVWw
kWdVha4s38XM/pa/yr47av7+z3VTmvDRyAHcaT92whREFpLv9cj5lTeJSibyr/Mr
m/YtjCZVWgaOYIhwrXwKLqPr/11inWsAkfIytvHWTxZYEcXLgAXFuUuaS3uF9gEi
NQwzGTU1v0FqkqTBr4B8nW3HCN47XUu0t8Y0e+lf4s4OxQawWD79J9/5d3Ry0vbV
3Am1FtGJiJvOwRsIfVChDpYStTcHTCMqtvWbV6L11BWkpzGXSW4Hv43qa+GSYOD2
QU68Mb59oSk2OB+BtOLpJofmbGEGgvmwyCI9MwIDAQABAoIBACiARq2wkltjtcjs
kFvZ7w1JAORHbEufEO1Eu27zOIlqbgyAcAl7q+/1bip4Z/x1IVES84/yTaM8p0go
amMhvgry/mS8vNi1BN2SAZEnb/7xSxbflb70bX9RHLJqKnp5GZe2jexw+wyXlwaM
+bclUCrh9e1ltH7IvUrRrQnFJfh+is1fRon9Co9Li0GwoN0x0byrrngU8Ak3Y6D9
D8GjQA4Elm94ST3izJv8iCOLSDBmzsPsXfcCUZfmTfZ5DbUDMbMxRnSo3nQeoKGC
0Lj9FkWcfmLcpGlSXTO+Ww1L7EGq+PT3NtRae1FZPwjddQ1/4V905kyQFLamAA5Y
lSpE2wkCgYEAy1OPLQcZt4NQnQzPz2SBJqQN2P5u3vXl+zNVKP8w4eBv0vWuJJF+
hkGNnSxXQrTkvDOIUddSKOzHHgSg4nY6K02ecyT0PPm/UZvtRpWrnBjcEVtHEJNp
bU9pLD5iZ0J9sbzPU/LxPmuAP2Bs8JmTn6aFRspFrP7W0s1Nmk2jsm0CgYEAyH0X
+jpoqxj4efZfkUrg5GbSEhf+dZglf0tTOA5bVg8IYwtmNk/pniLG/zI7c+GlTc9B
BwfMr59EzBq/eFMI7+LgXaVUsM/sS4Ry+yeK6SJx/otIMWtDfqxsLD8CPMCRvecC
2Pip4uSgrl0MOebl9XKp57GoaUWRWRHqwV4Y6h8CgYAZhI4mh4qZtnhKjY4TKDjx
QYufXSdLAi9v3FxmvchDwOgn4L+PRVdMwDNms2bsL0m5uPn104EzM6w1vzz1zwKz
5pTpPI0OjgWN13Tq8+PKvm/4Ga2MjgOgPWQkslulO/oMcXbPwWC3hcRdr9tcQtn9
Imf9n2spL/6EDFId+Hp/7QKBgAqlWdiXsWckdE1Fn91/NGHsc8syKvjjk1onDcw0
NvVi5vcba9oGdElJX3e9mxqUKMrw7msJJv1MX8LWyMQC5L6YNYHDfbPF1q5L4i8j
8mRex97UVokJQRRA452V2vCO6S5ETgpnad36de3MUxHgCOX3qL382Qx9/THVmbma
3YfRAoGAUxL/Eu5yvMK8SAt/dJK6FedngcM3JEFNplmtLYVLWhkIlNRGDwkg3I5K
y18Ae9n7dHVueyslrb6weq7dTkYDi3iOYRW8HRkIQh06wEdbxt0shTzAJvvCQfrB
jg/3747WSsf/zBTcHihTRBdAv6OmdhV4/dD5YBfLAkLrd+mX7iE=
-----END RSA PRIVATE KEY-----'''
#   "sub": "1234567890",
#   "name":{
#      "first_name": "New",
#     "last_name": "Tester"
#   },
#   "email":"test@social.com",
#   "iat": 1516239022
MOCKED_USER_JWT = 'eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6eyJmaXJzdF9uYW1lIjoiTmV3IiwibGFzdF9uYW1lIjoiVGVzdGVyIn0sImVtYWlsIjoidGVzdEBzb2NpYWwuY29tIiwiaWF0IjoxNTE2MjM5MDIyfQ.EEPLjGIfOrcmQOrp3cdJZTybo7EM-sLU90LxFUSI8AAnzMBW-9P1wMmz8qA0njWXh_N7wuwKmtxTQWExyCjq1xKYRIep4Jmi8WoWenTriobgFima-J6XT92pef2ojVoRZ9nDM7NkLnIUzIHeE16YCrLKlr8glXeKIOy3hPqUC1HSAVsxoVvgur-FN4RoL3_CtWBkvSt1rKIg-bon1zII-a5_bA0VuKTkGrBPkTgN-qJNC5usCqlnUl7bzPptHlnsVI-Z9guy1ScpsZ1immppboFCYdkzr6tSIxpvsmNpuwGgR-neAYoY87MPB8o5LVsdUaSzi0PoeAcR-yevDCKUaw'
APPLE_MOCKED_test_user_JWT = 'eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6eyJmaXJzdF9uYW1lIjoiTmV3IiwibGFzdF9uYW1lIjoiVGVzdGVyIn0sImVtYWlsIjoidGVzdEBleGFtcGxlLmNvbSIsImlhdCI6MTUxNjIzOTAyMn0.X4bUa9YfxDv9HGV_2bTJBFasvZegshlZBwCJuBzOfaCda5z3t9D79F3fBSA-PXsBxlvFczm-DrmbVjhoei5RKgpmZK-USdAodjxNVa5BTr1FkQkLKvhRS492sygOk0ZpX7s3Holk3fVURZ6djn0mH2SfvqUf2hAi7YOunWmaf6oFkJr2Bsq4Nszzcx4EwFNOqYloH4LRPQX60lIIN7nS0mWzwlaZFl01xhwlMg11EorRp6zPQnr_XXlNsyarFeV_uD6jEcQOXs2sT15hRseDHiTGBGkHPsuKY-pAcs_nSpsTjS0L1gIiMOeH1UdkDFEjbczuXx-7pkvOh6KBqKSc5g'
APPLE_MOCKED_JWKS = {
    "keys": [
        {
            "alg": "RS256",
            "use": "sig",
            "kid": "zFHVsM9dc4OlHIwvEnVqfKzRj1ujqYGsZnXAcgn_CqI",
            "e": "AQAB",
            "kty": "RSA",
            "n": 'nzyis1ZjfNB0bBgKFMSvvkTtwlvBsaJq7S5wA-kzeVOVpVWwkWdVha4s38XM_pa_yr47av7-z3VTmvDRyAHcaT92whREFpLv9c'
                 'j5lTeJSibyr_Mrm_YtjCZVWgaOYIhwrXwKLqPr_11inWsAkfIytvHWTxZYEcXLgAXFuUuaS3uF9gEiNQwzGTU1v0FqkqTBr4B8'
                 'nW3HCN47XUu0t8Y0e-lf4s4OxQawWD79J9_5d3Ry0vbV3Am1FtGJiJvOwRsIfVChDpYStTcHTCMqtvWbV6L11BWkpzGXSW4Hv4'
                 '3qa-GSYOD2QU68Mb59oSk2OB-BtOLpJofmbGEGgvmwyCI9Mw'
        }
    ]
}

FACEBOOK_MOCKED_USERINFO = {
    "id": "624123456789000", "name": "Tester Fb", "email": "test@social.com",
    "picture": {"data":
                    {"url": "https://platform-lookaside.fbsbx.com/",
                     "width": 50, "height": 50,
                    }
                },
    "last_name": "Fb",
    "first_name": "Tester"
}

def mock_requests_factory(provider: str) -> Mock:
    json_return_value = {
        'apple': APPLE_MOCKED_JWKS,
        'facebook': FACEBOOK_MOCKED_USERINFO,
        'google': APPLE_MOCKED_JWKS
    }
    return Mock(**{
        'json.return_value': json_return_value.get(provider),
        'content': json.dumps(json_return_value.get(provider)),
        'status_code': 200,
        'ok': True
    })


class UserSignupTestSerializer(serializers.Serializer):
    first_name = serializers.CharField(allow_blank=True)
    last_name = serializers.CharField()

    def validate_first_name(self, first_name):
        return 'Override'


class SocialTest(BaseTest):

    def _hard_reload_serializer_mod(self):
        from importlib import reload
        from auth_framework.social import serializers
        from auth_framework.serializers import signup_serializers
        reload(signup_serializers)
        reload(serializers)

    @patch('auth_framework.social.providers.abstract_adapter.requests.get', Mock(
        side_effect=lambda k: {'https://appleid.apple.com/auth/keys': mock_requests_factory('apple'),}.get(k)))
    def test_apple_new_user(self):
        response = self.client.post(reverse('apple_login'), data={'user_jwt': MOCKED_USER_JWT})
        self.assertEqual(response.status_code, 406)
        json_resp = response.json()
        self.assertEqual(json_resp['non_field_error']['message'], 'Auto SignUp disabled.')
        self.assertEqual(json_resp['non_field_error']['social_login'],
                         '{"account": {"uid": "1234567890", "provider": "apple", '
                         '"extra_data": {"sub": "1234567890", "name": {"first_name": "New", "last_name": "Tester"}, '
                         '"email": "test@social.com", "iat": 1516239022}},'
                         ' "user": {"email": "test@social.com", "first_name": "New", "last_name": "Tester"}}'
                         )

    @patch('auth_framework.social.providers.abstract_adapter.requests.get', Mock(
        side_effect=lambda k: {'https://appleid.apple.com/auth/keys': mock_requests_factory('apple'), }.get(
            k)))
    def test_apple_existed_email(self):
        response = self.client.post(reverse('apple_login'), data={'user_jwt': APPLE_MOCKED_test_user_JWT})
        self.assertEqual(response.status_code, 400)
        json_resp = response.json()
        self.assertIn('email', json_resp)
        self.assertEqual(response.content.find(b'social_login'), -1)

    @patch('auth_framework.social.providers.abstract_adapter.requests.get', Mock(
        side_effect=lambda k: {'https://appleid.apple.com/auth/keys': mock_requests_factory('apple'), }.get(
            k)))
    def test_apple_existed_account(self):
        """
        if uid existed return access_token as token api
        """
        SocialAccountModel.objects.create(user=self.test_user, provider='apple', uid='1234567890')
        response = self.client.post(reverse("apple_login"), data={'user_jwt': APPLE_MOCKED_test_user_JWT})
        self.assertEqual(response.status_code, 200)
        content = response.data
        self.assertIn('access_token', content)
        self.assertIn('refresh_token', content)

    def test_create_social(self):
        url = reverse('create-user')
        data = {'email': 'test@social.com', 'password1': '1234qwes', 'password2': '1234qwes',
                'social_login': '{"account": {"uid": "1234567890", "provider": "apple", '
                                '"extra_data": {"sub": "1234567890",'
                                ' "name": {"first_name": "New", "last_name": "Tester"}, '
                                '"email": "test@social.com", "iat": 1516239022}},'
                                ' "user": {"email": "test@social.com", "first_name": "New", "last_name": "Tester"}}'
                 }
        response = self.client.post(url, data,)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(UserModel.objects.filter(email='test@social.com').exists())
        self.assertTrue(SocialAccountModel.objects.filter(uid='1234567890', provider='apple').exists())

    @patch('auth_framework.social.providers.abstract_adapter.requests.get', Mock(
        side_effect=lambda k: {'https://appleid.apple.com/auth/keys': mock_requests_factory('apple'), }.get(
            k)))
    def test_create_social_w_custom_email(self):
        """
        if email was not granted or user using alt email to create account
        """
        url = reverse('create-user')
        data = {'email': 'test_alt@social.com', 'password1': '1234qwes', 'password2': '1234qwes',
                'social_login': '{"account": {"uid": "1234567890", "provider": "apple", '
                                '"extra_data": {"sub": "1234567890",'
                                ' "name": {"first_name": "New", "last_name": "Tester"}, "iat": 1516239022}},'
                                ' "user": {"email": null, "first_name": "New", "last_name": "Tester"}}'
                 }
        response = self.client.post(url, data,)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(UserModel.objects.filter(email='test_alt@social.com').exists())
        self.assertTrue(SocialAccountModel.objects.filter(uid='1234567890', provider='apple').exists())
        response = self.client.post(reverse("apple_login"), data={'user_jwt': MOCKED_USER_JWT})
        self.assertEqual(response.status_code, 200)
        self.assertIn('access_token', response.data)

    @patch('auth_framework.social.providers.facebook.adapter.requests.get', Mock(
        side_effect=lambda k, params: {GRAPH_API_URL + "/me": mock_requests_factory('facebook'), }.get(k)))
    def test_fb_new_user(self):
        response = self.client.post(reverse('fb_login'), data={'access_token': 'facebook_fake_access_token'})
        self.assertEqual(response.status_code, 406)
        json_resp = response.json()
        self.assertEqual(json_resp['non_field_error']['message'], 'Auto SignUp disabled.')
        self.assertEqual(json_resp['non_field_error']['social_login'],
                         '{"account": {"uid": "624123456789000", "provider": "facebook", '
                         '"extra_data": {"id": "624123456789000", "name": "Tester Fb",'
                         ' "email": "test@social.com", "picture": {"data": {"url": '
                         '"https://platform-lookaside.fbsbx.com/", "width": 50, "height": 50}}, '
                         '"last_name": "Fb", "first_name": "Tester"}},'
                         ' "user": {"email": "test@social.com", "first_name": "Tester", "last_name": "Fb"}}'
                         )

    @patch('auth_framework.social.providers.abstract_adapter.requests.get', Mock(
        side_effect=lambda k: {'https://www.googleapis.com/oauth2/v3/certs': mock_requests_factory('google'), }.get(k)))
    def test_gl_new_user(self):
        response = self.client.post(reverse('gl_login'), data={'user_jwt': MOCKED_USER_JWT})
        self.assertEqual(response.status_code, 406)
        json_resp = response.json()
        self.assertEqual(json_resp['non_field_error']['message'], 'Auto SignUp disabled.')
        self.assertEqual(json_resp['non_field_error']['social_login'],
                         '{"account": {"uid": "1234567890", "provider": "google", '
                         '"extra_data": {"sub": "1234567890", "name": {"first_name": "New", "last_name": "Tester"}, '
                         '"email": "test@social.com", "iat": 1516239022}},'
                         ' "user": {"email": "test@social.com", "first_name": "", "last_name": ""}}'
                         )

    @override_settings(AUTH_FRAMEWORK={"SOCIAL_AUTO_SIGNUP": True,})
    @patch('auth_framework.social.providers.abstract_adapter.requests.get', Mock(
        side_effect=lambda k: {'https://www.googleapis.com/oauth2/v3/certs': mock_requests_factory('google'), }.get(k)))
    def test_auto_signup(self):
        response = self.client.post(reverse('gl_login'), data={'user_jwt': MOCKED_USER_JWT})
        self.assertEqual(response.status_code, 200)
        json_resp = response.json()
        self.assertTrue(UserModel.objects.filter(email="test@social.com").exists())
        user = UserModel.objects.get(email="test@social.com")
        self.assertEqual(user.has_usable_password(), False)
        self.assertIn('access_token', json_resp)

    @override_settings(AUTH_FRAMEWORK={"SOCIAL_AUTO_SIGNUP": True, "SERIALIZERS": {
        'SIGNUP_SERIALIZER': UserSignupTestSerializer}})
    @patch('auth_framework.social.providers.abstract_adapter.requests.get', Mock(
        side_effect=lambda k: {'https://www.googleapis.com/oauth2/v3/certs': mock_requests_factory('google'),
                               'https://appleid.apple.com/auth/keys': mock_requests_factory('apple'),
                               }.get(k)))
    def test_custom_validation_auto_signup(self):
        self._hard_reload_serializer_mod()
        response = self.client.post(reverse('gl_login'), data={'user_jwt': MOCKED_USER_JWT})
        self.assertEqual(response.status_code, 400)
        response = self.client.post(reverse('apple_login'), data={'user_jwt': MOCKED_USER_JWT})
        self.assertEqual(response.status_code, 200)
        json_resp = response.json()
        self.assertTrue(UserModel.objects.filter(email="test@social.com", first_name='Override').exists())


import json

from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.db import transaction, IntegrityError
from django.utils.translation import gettext_lazy as _
from requests import HTTPError
from rest_framework import serializers
from rest_framework.exceptions import AuthenticationFailed, NotAcceptable

from ..models import get_socialaccount_model
from ..serializers.signup_serializers import SignUpSerializer
from ..settings import app_settings
from .providers.abstract_adapter import OauthAdapterError, InvalidJWT

User = get_user_model()
SocialAccount = get_socialaccount_model()


class SocialSignUpSerializer(SignUpSerializer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from_auto_signup = self.context.get('auto_signup', False)
        if from_auto_signup:
            self.fields['password1'].required = False
            if 'password2' in self.fields:
                self.fields['password2'].required = False
        self.social_adapter = self.context.get('social_data', {})

    def populate_user(self):
        return self.social_adapter.user

    def save(self):
        with transaction.atomic():
            user = super().save()
            # TODO: reedit
            self.social_adapter.save(connect_user=user)
            return user


class SocialLoginSerializer(serializers.Serializer):
    access_token = serializers.CharField(required=False, allow_blank=True)
    code = serializers.CharField(required=False, allow_blank=True)
    user_jwt = serializers.CharField(required=False, allow_blank=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        view = self.context.get('view')
        self.request = self.context.get('request')
        self.adapter_class = getattr(view, 'adapter_class')

    def complete_social_login(self, adapter):
        assert not adapter.is_existed
        adapter.get_current_social_acct()
        process = self.request.GET.get('process', 'login')
        if process == "connect":
            if self.request.user.is_anonymous:
                raise AuthenticationFailed
            if adapter.is_existed and adapter.user != self.request.user:
                    raise serializers.ValidationError(
                        {"non_field_error": {
                            'message': _('The social account is already connected to a different account.'),
                            'code': 'AOSL007',
                            }})
            else:
                adapter.save(connect_user=self.request.user)
            return adapter.social_acct
        else:
            if adapter.is_existed:
                pass  # TODO: call django_login() when required
            else:
                is_email_taken = User.objects.filter(email__iexact=adapter.user.email).exists()
                if not app_settings.SOCIAL_AUTO_SIGNUP:
                    # Since auto signup disabled, we dont have to preserve unique email,
                    # instead we return errors to frontend and waiting incoming request.
                    if adapter.user.email and is_email_taken:
                        raise serializers.ValidationError({'email':
                            {
                                'message': 'Social account email: %s already has an account with %s'
                                           % (adapter.user.email, get_current_site(self.request).name),
                                'code': 'AOSL001'
                            }
                        })
                    else:
                        serialized_sociallogin = adapter.serialize()
                        raise NotAcceptable(
                            {"non_field_error": {'message': _("Auto SignUp disabled."),
                                                 'code': 'AOSL002',
                                                 'social_login': json.dumps(serialized_sociallogin),
                                                 }}

                        )
                else:
                    # For Auto Signup, we have to assure at least email is granted and keep unique in db
                    if adapter.user.email and app_settings.UNIQUE_EMAIL:
                        if is_email_taken:
                            raise serializers.ValidationError(
                                {'description': 'Social account email: %s already has an account with %s'
                                                % (adapter.user.email, get_current_site(self.request).name),
                                 'email': adapter.user.email,
                                 'code': 'AOSL003'
                                 }
                            )
                        else:
                            social_signup_serializer = SocialSignUpSerializer(data=adapter.serialize()['user'],
                                                                              context={'social_data': adapter,
                                                                                       'auto_signup': True})
                            social_signup_serializer.is_valid(raise_exception=True)
                            social_signup_serializer.save()
                            # we perform a quick user creation
                            # social_acct = adapter.save()
                            # call django_login
                            # return social_acct
                    else:
                        serialized_socialdata = adapter.serialize()
                        raise serializers.ValidationError(
                            {'description': 'Please input email',
                             'social_login': json.dumps(serialized_socialdata),
                             'code': 'AOSL004'
                             }
                        )

    def validate(self, attrs):
        adapter = self.adapter_class()
        access_token = attrs.get('access_token', None)
        code = attrs.get('code', None)
        user_jwt = attrs.get('user_jwt', None)
        if access_token or user_jwt:
            pass
        # Case 2: We received the authorization code, though here is less likely called after google changed to JWT
        elif code:
            try:
                access_token = adapter.get_access_token(code)
            except OauthAdapterError as err:
                raise serializers.ValidationError(
                    {"code": {'message': err, 'code': 'AOSL001'}})
        else:
            raise serializers.ValidationError(
                {"non_field_error": {'message': _("Incorrect input. user_jwt or access_token or code is required."),
                                     'code': 'AOSL005'
                                     }})
        try:
            social_acct, user = adapter.populate_social_acct(access_token=access_token,user_jwt=user_jwt)

            self.complete_social_login(adapter)
        except (HTTPError, InvalidJWT):
            raise serializers.ValidationError(
                {"non_field_error": {'message': _("Credentials failed with the provider."),
                                     'code': 'AOSL006'
                                     }})
        except IntegrityError as err:
            raise serializers.ValidationError('Duplicated Call %s' % err)
        # if not login.is_existing:
        #     login.lookup()
        #     login.save(request, connect=True)
        attrs['user'] = adapter.user
        return attrs


class SocialConnectionSerializer(serializers.Serializer):
    account = serializers.CharField(required=True)

    def __init__(self, *args, **kwargs):
        self.request = kwargs.get('context',{}).get('request')
        self.accounts = SocialAccount.objects.filter(user=self.request.user)
        super().__init__(*args, **kwargs)

    def validate_account(self, value):
        for sc in self.accounts:
            if sc.provider == value:
                return sc
        raise serializers.ValidationError("account name is not one of the available choices.", code='ASD001')

    def validate(self, attrs):
        account = attrs.get('account')
        accounts = self.accounts
        if account:
            if len(accounts) == 1:
                # No usable password happens on single auto signup account
                if not account.user.has_usable_password():
                    raise serializers.ValidationError({"non_field_error": {
                        'message': _("Your account has no password set up."), 'code': 'ASD002',}})
        return attrs

    def save(self):
        account = self.validated_data['account']
        account.delete()



__author__ = 'Carrycat'
from django.contrib.auth import login as django_login, logout as django_logout
from rest_framework.authentication import SessionAuthentication
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView

from ..permissions import IsAuthenticatedAndActive
from ..serializers.account_serializers import SessionLoginSerializer
from ..settings import import_callable, app_settings

user_serializers = import_callable(app_settings.SERIALIZERS.get('USERINFO_SERIALIZER'))


class SessionLogin(GenericAPIView):
    permission_classes = (AllowAny,)
    serializer_class = SessionLoginSerializer
    validated_user = None

    def perform_login(self, user):
        django_login(self.request, user, backend='django.contrib.auth.backends.ModelBackend')

    def post(self, *args, **kwargs):
        login_serializer = self.get_serializer(data=self.request.data)
        login_serializer.is_valid(raise_exception=True)
        self.validated_user = login_serializer.validated_data
        self.perform_login(self.validated_user)
        data = user_serializers(instance=self.validated_user).data if user_serializers else {}
        return Response(data)


class SessionLogout(APIView):
    permission_classes = (IsAuthenticatedAndActive,)
    authentication_classes = (SessionAuthentication,)
    # authentication_classes = (CsrfExemptSessionAuthentication,)

    def perform_logout(self):
        django_logout(self.request)

    def post(self, request, *args, **kwargs):
        self.perform_logout()
        return Response({})



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as DjangoUserAdmin
from django.contrib.auth.forms import UserChangeForm
from django.forms import CharField

from .models import get_socialaccount_model, User
from .settings import app_settings, import_callable

# Register your models here.
default_fields = ['username', 'email', 'first_name', 'last_name']
default_list_play = ['id',  'email', 'first_name', 'last_name']
default_search_fieldsets = ['email']
default_add_fieldsets = ['username', 'email', 'password1', 'password2']

SocialAccount = get_socialaccount_model()

if app_settings.USE_PHONENUMBER_FIELD:
    from phonenumber_field.formfields import PhoneNumberField
    default_fields.append('phone_number')
    default_list_play.append('phone_number')
    default_search_fieldsets.append('phone_number')
    default_add_fieldsets.insert(0,'phone_number')
else:
    class PhoneNumberField(CharField):
        pass


class UserChangeForm(UserChangeForm):
    phone_number = PhoneNumberField()

    class Meta:
        model = get_user_model()
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        if not app_settings.USE_PHONENUMBER_FIELD:
            self.declared_fields.pop('phone_number')
            self.base_fields.pop('phone_number')
        super().__init__(*args, **kwargs)


class UserAdmin(DjangoUserAdmin):
    list_display = tuple(default_list_play)
    fieldsets = (
        ('Personal info', {'fields': default_fields}),
        ('Extras', {'fields': ('id', 'password')}),
        ('Permissions', {'fields': ('is_active', 'is_staff', 'is_superuser'), 'classes': ['collapse']}),
    )
    search_fieldsets = tuple(default_search_fieldsets)
    add_fieldsets = (
        (None, {'fields': tuple(default_add_fieldsets)}),
        ('Personal info', {'fields': ('first_name', 'last_name',)}),
        ('Permissions', {'fields': ('is_active', 'is_staff', 'is_superuser'), 'classes': ['collapse']}),
    )
    form = UserChangeForm
    readonly_fields = ('id', 'last_login', 'date_joined')


class SocialAccountAdmin(admin.ModelAdmin):
    search_fields = []
    raw_id_fields = ("user",)
    list_display = ("user", "uid", "provider")
    list_filter = ("provider",)


admin.site.register(User, UserAdmin)
admin.site.register(SocialAccount, import_callable(app_settings.SOCIALACCOUNT_ADMIN_CLASS))



from django.apps import apps
from django.db import models
from django.utils.translation import gettext_lazy as _

from .settings import app_settings, import_callable
from .abstract_models import DefaultAbstractUser, PhoneNumberMixin, AvatarMixin, AbstractSocialAccount


def base_class():
    bases = [DefaultAbstractUser,]
    if app_settings.USE_PHONENUMBER_FIELD:
        bases.append(PhoneNumberMixin)
    return tuple(bases)


def settings_default_value():
    return {
            "language": "en",
            "reminder": {
                "interval": 15,
                "sleep": 1260,
                "wakeUp": 540
            },
            "unit": "metric"
    }


class AbstractUser(*base_class()):
    class Meta(DefaultAbstractUser.Meta):
        abstract = True


class User(*base_class()):
    # settings = models.JSONField(_('account settings'), default=dict)

    class Meta(DefaultAbstractUser.Meta):
        verbose_name = _('user')
        verbose_name_plural = _('users')
        swappable = 'AUTH_USER_MODEL'

    @classmethod
    def get_fields(cls):
        return tuple(set(map(lambda field: field.name, cls._meta.local_fields)) -
                     {'password', 'is_superuser', 'is_active', 'last_login', 'is_staff', 'date_joined'})


class SocialAccount(AbstractSocialAccount):

    class Meta(AbstractSocialAccount.Meta):
        swappable = "REST_AUTH_SOCIAL_ACCOUNT"


def get_socialaccount_model():
    """ Return the RefreshToken model that is active in this project. """
    return apps.get_model(app_settings.SOCIALACCOUNT_MODEL)


def get_socialaccount_admin_class():
    """ Return the Application admin class that is active in this project. """
    socialaccount_admin_class = app_settings.SOCIALACCOUNT_ADMIN_CLASS
    return import_callable(socialaccount_admin_class)


from django.conf import settings
from django.utils import timezone
from oauth2_provider.models import Application, AccessToken, RefreshToken
from oauth2_provider.settings import oauth2_settings
from oauthlib.common import generate_token


DEFAULT_SCOPES = getattr(settings, 'OAUTH_DEFAULT_SCOPES', None)
DEFAULT_USER_SCOPES = getattr(settings, 'OAUTH_DEFAULT_USER_SCOPES', None)


# Used for Social Login and Sign Up, never validate the user, the expires date.
def get_access_token(user, request=None):
    """
    Takes a user instance and return an access_token as a JsonResponse
    instance.
    """
    validator_class = oauth2_settings.OAUTH2_VALIDATOR_CLASS
    # our oauth2 app
    validator = validator_class()


    # We delete the old access_token and refresh_token  # TODO: we just ignore old access token
    # try:
    #     old_access_token = AccessToken.objects.get(
    #         user=user, application=app)
    #     old_refresh_token = RefreshToken.objects.get(
    #         user=user, access_token=old_access_token
    #     )
    # except:
    #     pass
    # else:
    #     old_access_token.delete()
    #     old_refresh_token.delete()

    token = {
        'access_token': generate_token(),
        'refresh_token': generate_token(),
        'scope': ' '.join(oauth2_settings._DEFAULT_SCOPES),
        'expires_in':  oauth2_settings.ACCESS_TOKEN_EXPIRE_SECONDS
    }
    request.user = user
    request.scopes = token['scope'].strip().split(" ")
    request.grant_type = request.client.authorization_grant_type
    validator.save_token(token, request, nonce='server-gen')
    # we create the access token
    return token


# def import_callable(path_or_callable):
#     if hasattr(path_or_callable, '__call__'):
#         return path_or_callable
#     else:
#         assert isinstance(path_or_callable, str)
#         package, attr = path_or_callable.rsplit('.', 1)
#         return getattr(import_module(package), attr)


# def complete_social_login(request, sociallogin):
#     assert not sociallogin.is_existing
#     sociallogin.lookup()
#     try:
#         social_get_adapter().pre_social_login(request, sociallogin)
#         social_account_signals.pre_social_login.send(sender=SocialLogin,
#                                       request=request,
#                                       sociallogin=sociallogin)
#     except ImmediateHttpResponse as e:
#         return e.response
#     process = sociallogin.state.get('process')
#     if process == AuthProcess.REDIRECT:
#         return _social_login_redirect(request, sociallogin)
#     elif process == AuthProcess.CONNECT:
#         return _add_social_account(request, sociallogin)
#     else:
#         return _complete_social_login(request, sociallogin)
#
#
# def _process_signup(request, sociallogin):
#
#     if not social_app_settings.AUTO_SIGNUP:
#         if sociallogin.user.email and email_address_exists(sociallogin.user.email):
#             raise serializers.ValidationError({'email':
#                                                    {
#                                                        'message': 'Social account email: %s already has an account with %s'
#                                                                   % (sociallogin.user.email, get_current_site(request))}})
#         else:
#             serialized_sociallogin = sociallogin.serialize()
#             raise serializers.ValidationError(
#                 {'message': 'Auto SignUp disabled',
#                  'social_login': json.dumps(serialized_sociallogin)
#                  }
#             )
#     else:
#         auto_signup = social_get_adapter().is_auto_signup_allowed(request,
#                                                            sociallogin)
#         if not auto_signup:
#             if sociallogin.user.email:
#                 raise serializers.ValidationError(
#                     {'description': 'Social account email: %s already has an account with %s'
#                                     % (sociallogin.user.email, get_current_site(request)),
#                      'email': sociallogin.user.email
#                      }
#                 )
#             else:  # we should reject this request since no email input
#                 serialized_sociallogin = sociallogin.serialize()
#                 raise serializers.ValidationError(
#                     {'description': 'Please input email',
#                      'social_login': json.dumps(serialized_sociallogin)
#                      }
#                 )
#         else:
#             # Ok, auto signup it is, at least the e-mail address is ok.
#             # We still need to check the username though...
#             if account_settings.USER_MODEL_USERNAME_FIELD:
#                 username = user_username(sociallogin.user)
#                 try:
#                     get_adapter().clean_username(username)
#                 except ValidationError:
#                     # This username is no good ...
#                     user_username(sociallogin.user, '')
#             # FIXME: This part contains a lot of duplication of logic
#             # ("closed" rendering, create user, send email, in active
#             # etc..)
#             try:
#                 if not social_get_adapter().is_open_for_signup(request,
#                                                         sociallogin):
#                     # return render(request,
#                     #               "account/signup_closed.html")
#                     pass
#                     # FIXME: Shouldn't go here
#             except ImmediateHttpResponse as e:
#                 return e.response
#             social_get_adapter().save_user(request, sociallogin, form=None)
#             ret = complete_signup(request, sociallogin.user,
#                                   'none',
#                                   sociallogin.get_redirect_url(request),
#                                   signal_kwargs={'sociallogin': sociallogin})
#         return ret
#
#
# def _complete_social_login(request, sociallogin):
#     # if request.user.is_authenticated():
#         # get_account_adapter().logout(request)
#         # FIXME: Shouldn't go here. Request from facebook may be user = None !!!
#         # Since General Acrossor user is None.
#         # pass
#     if sociallogin.is_existing:
#         # Login existing user
#         ret = perform_login(request, sociallogin.user, email_verification=social_app_settings.EMAIL_VERIFICATION,
#                             redirect_url=sociallogin.get_redirect_url(request),
#                             signal_kwargs={"sociallogin": sociallogin})
#     else:
#         # New social user
#         ret = _process_signup(request, sociallogin)
#     return ret
#
#
# def _add_social_account(request, sociallogin):
#     if request.user.is_anonymous():
#         # This should not happen. Simply redirect to the connections
#         # view (which has a login required)
#         # return HttpResponseRedirect(reverse('socialaccount_connections'))
#         raise AuthenticationFailed
#     if sociallogin.is_existing:
#         if sociallogin.user != request.user:
#             # Social account of other user. For now, this scenario
#             # is not supported. Issue is that one cannot simply
#             # remove the social account from the other user, as
#             # that may render the account unusable.
#             raise serializers.ValidationError(
#                 {'description': 'The social account is already connected to a different account.'
#                  })
#         else:
#             # This account is already connected -- let's play along
#             # and render the standard "account connected" message
#             # without actually doing anything.
#             pass
#     else:
#         # New account, let's connect
#         sociallogin.connect(request, request.user)
#         try:
#             social_account_signals.social_account_added.send(sender=SocialLogin,
#                                               request=request,
#                                               sociallogin=sociallogin)
#         except ImmediateHttpResponse as e:
#             return e.response
#     # default_next = social_get_adapter() \
#     #     .get_connect_redirect_url(request,
#     #                               sociallogin.account)
#     # next_url = sociallogin.get_redirect_url(request) or default_next
#     # return HttpResponseRedirect(next_url)
#
#
# def perform_login(request, user, email_verification,
#                   redirect_url=None, signal_kwargs=None,
#                   signup=False):
#     """
#     Keyword arguments:
#
#     signup -- Indicates whether or not sending the
#     email is essential (during signup), or if it can be skipped (e.g. in
#     case email verification is optional and we are only logging in).
#     """
#     # Local users are stopped due to form validation checking
#     # is_active, yet, adapter methods could toy with is_active in a
#     # `user_signed_up` signal. Furthermore, social users should be
#     # stopped anyway.
#     if not user.is_active:
#         raise serializers.ValidationError({'description': 'User account is disabled.'})
#
#     from allauth.account.models import EmailAddress
#     has_verified_email = EmailAddress.objects.filter(user=user,
#                                                      verified=True).exists()
#     if email_verification == account_settings.EmailVerificationMethod.NONE:
#         pass
#     elif email_verification == account_settings.EmailVerificationMethod.OPTIONAL:
#         # In case of OPTIONAL verification: send on signup.
#         if not has_verified_email and signup:
#             send_email_confirmation(request, user, signup=signup)
#     elif email_verification == account_settings.EmailVerificationMethod.MANDATORY:
#         if not has_verified_email:
#             send_email_confirmation(request, user, signup=signup)
#             return HttpResponseRedirect(
#                 reverse('account_email_verification_sent'))
#     try:
#         get_adapter().login(request, user)
#         response = HttpResponseRedirect('/')
#         if signal_kwargs is None:
#             signal_kwargs = {}
#         signals.user_logged_in.send(sender=user.__class__,
#                                     request=request,
#                                     response=response,
#                                     user=user,
#                                     **signal_kwargs)
#         get_adapter().add_message(request,
#                                   messages.SUCCESS,
#                                   'account/messages/logged_in.txt',
#                                   {'user': user})
#     except ImmediateHttpResponse as e:
#         response = e.response
#     return response
#
#
# def complete_signup(request, user, email_verification, success_url,
#                     signal_kwargs=None):
#     if signal_kwargs is None:
#         signal_kwargs = {}
#     signals.user_signed_up.send(sender=user.__class__,
#                                 request=request,
#                                 user=user,
#                                 **signal_kwargs)
#     return perform_login(request, user,
#                          email_verification=email_verification,
#                          signup=True,
#                          redirect_url=success_url,
#                          signal_kwargs=signal_kwargs)
#
#
#
#
# def custom_exception_handler(exc, context):
#     """
#     Returns the response that should be used for any given exception.
#
#     By default we handle the REST framework `APIException`, and also
#     Django's built-in `Http404` and `PermissionDenied` exceptions.
#
#     Any unhandled exceptions may return `None`, which will cause a 500 error
#     to be raised.
#     """
#     if isinstance(exc, exceptions.APIException):
#         headers = {}
#         if getattr(exc, 'auth_header', None):
#             headers['WWW-Authenticate'] = exc.auth_header
#         if getattr(exc, 'wait', None):
#             headers['Retry-After'] = '%d' % exc.wait
#
#         if isinstance(exc.detail, (list, dict)):
#             data = exc.detail
#             # for key, value in exc.detail.items():
#             #     data[key] = value[0]
#         else:
#             data = {'detail': exc.detail}
#
#         set_rollback()
#         return Response(data, status=exc.status_code, headers=headers)
#
#     elif isinstance(exc, Http404):
#         msg = _('Not found.')
#         data = {'detail': str(msg)}
#
#         set_rollback()
#         return Response(data, status=status.HTTP_404_NOT_FOUND)
#
#     elif isinstance(exc, PermissionDenied):
#         msg = _('Permission denied.')
#         data = {'detail': str(msg)}
#
#         set_rollback()
#         return Response(data, status=status.HTTP_403_FORBIDDEN)
#
#     # Note: Unhandled exceptions will raise a 500 error.
#     return None



from importlib import import_module

from django.conf import settings
from django.core.signals import setting_changed

SOCIALACCOUNT_MODEL = getattr(settings, "REST_AUTH_SOCIALACCOUNT_MODEL", "auth_framework.SocialAccount")

DEFAULTS = {
    'UNIQUE_EMAIL': True,
    'RESET_PASSWORD_BY': 'pin',  # 'url'| 'pin'
    'DEBUG_PIN': None,
    'TITLE_IN_MESSAGE': None,
    'SERIALIZERS': {
        # 'SOCIAL_LOGIN_SERIALIZER': 'auth.social.serializers.DefaultSocialLoginSerializer',
        'SIGNUP_SERIALIZER': 'auth_framework.serializers.signup_serializers.DefaultSignUpSerializer',
        'USERINFO_SERIALIZER': None
    },
    'SOCIALACCOUNT_MODEL': SOCIALACCOUNT_MODEL,
    'SOCIALACCOUNT_ADMIN_CLASS': "auth_framework.admin.SocialAccountAdmin",
    # SOCIAL LOGINS
    'SOCIAL_CALLBACK_URL': None,  # eg: 'https://developers.google.com/oauthplayground'
    'SOCIAL_AUTO_SIGNUP': False,
    # SIGN UP
    # 'SIGNUP_EMAIL_VERIFICATION': 'none', # trimmed out email verification celery task in closed source. fewer usage
    'SIGNUP_USERNAME_REQUIRED': False,
    'SIGNUP_USERNAME_VALIDATORS': [],
    'USE_PASSWORD_TWICE_VALIDATION': True,
    # ADVANCES
    'USE_PHONENUMBER_FIELD': False,
    'USE_CELERY_EMAIL': False,
    'USE_ID_TOKEN': True,
    'OAUTH_SAVE_ID_TOKEN': False
}


def import_callable(path_or_callable):
    if path_or_callable is None:
        return None
    if hasattr(path_or_callable, '__call__'):
        return path_or_callable
    else:
        assert isinstance(path_or_callable, str)
        package, attr = path_or_callable.rsplit('.', 1)
        return getattr(import_module(package), attr)


class AuthSettings:
    """
    """
    def __init__(self, user_settings=None, defaults=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'AUTH_FRAMEWORK', {})
        return self._user_settings

    @property
    def username_validators(self):
        from django.core.exceptions import ImproperlyConfigured
        from django.contrib.auth import get_user_model
        validators = self.user_settings.get("SIGNUP_USERNAME_VALIDATORS", None)
        if validators:
            ret = []
            if not isinstance(validators, list):
                raise ImproperlyConfigured(
                    "SIGNUP_USERNAME_VALIDATORS is expected to be a list"
                )
            for path in validators:
                pkg, attr = path.rsplit(".", 1)
                validator = getattr(import_module(pkg), attr)
                ret.append(validator())
        else:
            ret = (
                get_user_model()._meta.get_field('username').validators
            )
        return ret

    def serializers(self, data):
        # Check if present in user settings
        for key, value in data.items():
            data[key] = import_callable(value)
        return data

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
            if isinstance(val, dict):
                val = self.defaults[attr].copy()
                val.update(self.user_settings[attr])
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]
        if attr == 'SERIALIZERS':
            val = self.serializers(val)
        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


app_settings = AuthSettings(None, DEFAULTS)


def reload_app_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'AUTH_FRAMEWORK':
        app_settings.reload()


setting_changed.connect(reload_app_settings)



from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.cache import cache
from django.core.exceptions import ValidationError
from rest_framework import serializers
from django.utils.translation import gettext_lazy as _

try:
    from phonenumber_field.serializerfields import PhoneNumberField
except ImportError:
    class PhoneNumberField(serializers.CharField):
        pass
from ..settings import app_settings

User = get_user_model()


class EmailMixin(serializers.Serializer):
    email = serializers.EmailField()

    def __init__(self, *args, **kwargs):
        self.check_existed = getattr(self,'check_existed', None) or kwargs.pop('check_existed', False)
        super().__init__(*args, **kwargs)

    def validate_email(self, email):
        request_user = getattr(self.context.get('request', None), 'user', None)
        lookup_query = User.objects
        if request_user and request_user.pk:
            lookup_query = lookup_query.exclude(pk=request_user.pk)
        try:
            existed_user = lookup_query.get(email__iexact=email)
        except User.DoesNotExist:
            existed_user = None
        if self.check_existed:
            if existed_user:
                if hasattr(self, 'user_found'):
                    self.user_found = existed_user
                return email
            else:
                raise serializers.ValidationError(
                    _("The e-mail address is not assigned to any user account"),
                    code='AE001'
                )
        else:
            if app_settings.UNIQUE_EMAIL and existed_user:
                raise serializers.ValidationError(
                    _("A user is already registered with this e-mail address."),
                    code='AE002'
                )
            return email


class PasswordMixin(serializers.Serializer):
    password1 = serializers.CharField()
    password2 = serializers.CharField()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not app_settings.USE_PASSWORD_TWICE_VALIDATION:
            del self.fields["password2"]

    def validate_password1(self, password1):
        try:
            validate_password(password1, user=getattr(self,'user', None))
        except ValidationError as error:
            message = getattr(error, 'message', error.messages[0])
            raise serializers.ValidationError(message, code='AP001')
        return password1

    def validate(self, attrs):
        password = attrs.pop("password1", None)
        if (
                app_settings.USE_PASSWORD_TWICE_VALIDATION
                and "password2" in attrs
        ):
            if password != attrs.pop("password2", None):
                raise serializers.ValidationError({'password2': {
                    'message': _("You must type the same password each time."), 'code': 'AP002'}})

        attrs['password'] = password
        return attrs


class PhoneNumMixin(serializers.Serializer):
    phone_number = PhoneNumberField()

    def __init__(self, *args, **kwargs):
        self.skip_phone_validation = getattr(self, 'skip_phone_validation', False) or \
                                   kwargs.pop('skip_phone_validation', False)
        self.check_existed = getattr(self, 'check_existed', None) or kwargs.pop('check_existed', False)
        super().__init__(*args, **kwargs)

    def validate_phone_number(self, val):
        if self.skip_phone_validation:
            return val
        request_user = getattr(self.context.get('request', None), 'user', None)
        lookup_query = User.objects
        if request_user and request_user.pk:
            lookup_query = lookup_query.exclude(pk=request_user.pk)
        try:
            existed_user = lookup_query.get(phone_number=val)
        except User.DoesNotExist:
            existed_user = None
        if self.check_existed:
            if existed_user:
                if hasattr(self, 'user_found'):
                    self.user_found = existed_user
                return val
            else:
                raise serializers.ValidationError(
                    _("The phone number is not assigned to any user account"),
                    code='AP001'
                )
        else:
            if existed_user:
                raise serializers.ValidationError(_("This phone number was already taken."),
                                                  code='AP002')
            else:
                return val


class PhoneNumPinMixin(PhoneNumMixin):
    pin = serializers.IntegerField(max_value=999999)

    def __init__(self, *args, **kwargs):
        self.skip_pin_validation = getattr(self, 'skip_pin_validation', False) or\
                                   kwargs.pop('skip_pin_validation', False)
        super().__init__(*args, **kwargs)

    def validate(self, attrs):
        pin = attrs.get("pin")
        phone_number = attrs.get("phone_number")
        if self.skip_pin_validation:
            if not cache.get('is_verified:%s' % phone_number):
                raise serializers.ValidationError({'phone_number': {
                    'message': _("Phone is not verified."), 'code': 'AP004'}})
            cache.delete('is_verified:%s' % phone_number)
        elif phone_number and cache.get('pin_verify:%s' % phone_number) == pin or \
                app_settings.DEBUG_PIN and pin == app_settings.DEBUG_PIN:
            if not getattr(self, 'keep_alive', False):
                cache.delete('pin_verify:%s' % phone_number)
                cache.set('is_verified:%s' % phone_number, True, 60 * 10)
        elif phone_number:
            raise serializers.ValidationError({'pin': {
                'message': _("Invalid or expired PIN number."), 'code': 'AP003'}})
        return super().validate(attrs)




import json

import requests
from django.contrib.auth import get_user_model
from django.db import transaction
from jwcrypto import jwk, jwt
from jwcrypto.common import JWException
from jwcrypto.jwt import JWTExpired

from auth_framework.exceptions import AuthFrameworkAuthException
from auth_framework.models import get_socialaccount_model
from auth_framework.settings import app_settings

User = get_user_model()
SocialAccount = get_socialaccount_model()

# TODO: reedit

def general_deserializer(model, data):
    ret = model()
    for k, v in data.items():
        setattr(ret, k, v)
    return ret


class OauthAdapterError(AuthFrameworkAuthException):
    pass


class InvalidPubKey(OauthAdapterError):
    pass


class InvalidJWT(OauthAdapterError):
    pass


class OauthAdapter:
    access_token_url = None
    callback_url = app_settings.SOCIAL_CALLBACK_URL
    public_key_url = None
    access_token_method = "POST"
    client_id = ''
    client_secret = ''
    id = None

    def __init__(self):
        self.user = User()
        self.social_acct = SocialAccount()

    @property
    def is_existed(self) -> bool:
        """
        New social acct to be stored in db.
        """
        return self.social_acct.pk is not None

    def get_current_social_acct(self, update=False) -> None:
        """
        get social account in db.
        """
        assert not self.is_existed
        try:
            social_acct = SocialAccount.objects.get(
                provider=self.id, uid=self.social_acct.uid
            )
            social_acct.extra_data = self.social_acct.extra_data
            self.social_acct = social_acct
            self.user = self.social_acct.user
            if update:
                social_acct.save()
        except SocialAccount.DoesNotExist:
            pass

    def populate_social_acct(self, **kwargs) -> tuple[SocialAccount, User]:
        """
        """
        uid, model_fields, extra_data = self.get_uid(kwargs)

        self.social_acct = SocialAccount(extra_data=extra_data, uid=uid, provider=self.id)
        self.populate_user(model_fields)
        return self.social_acct, self.user

    def populate_user(self, data: dict[str:str]) -> None:
        for field, value in data.items():
            setattr(self.user, field, value)
        return self.user

    def get_uid(self, data: dict[str:str]) -> tuple[str, dict[str:str], dict[str:str]]:
        user_jwt = data.get('user_jwt', None)
        key = self.get_pubkey()
        try:
            jwt_token = jwt.JWT(key=key, jwt=user_jwt)
            claims = json.loads(jwt_token.claims)
            return claims['sub'], self.get_model_fields(claims), claims
        except (JWException, JWTExpired):
            raise InvalidJWT("invalid user_jwt")

    def get_model_fields(self, data):
        raise NotImplementedError(
            "The provider must implement the `extract_uid()` method"
        )

    def get_pubkey(self) -> jwk.JWKSet:
        resp = requests.get(self.public_key_url)
        resp.raise_for_status()
        try:
            jwks = jwk.JWKSet()
            jwks.import_keyset(resp.content)
            return jwks
        except jwk.InvalidJWKValue:
            raise InvalidPubKey("failed to get pubkey")

    def get_access_token(self, code) -> str:
        data = {
            "redirect_uri": self.callback_url,
            "grant_type": "authorization_code",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": code,
        }
        params = None
        url = self.access_token_url
        if self.access_token_method == "GET":
            params = data
            data = None
        resp = requests.request(
            self.access_token_method,
            url,
            params=params,
            data=data,
        )

        access_token = None
        if resp.status_code in [200, 201]:
            access_token = resp.json()
        if not access_token or "access_token" not in access_token:
            raise OauthAdapterError("Error retrieving access token: %s" % resp.json())
        return access_token['access_token']

    def save(self, connect_user: User = None) -> SocialAccount:
        assert not self.is_existed
        with transaction.atomic():
            if connect_user:
                self.user = connect_user
                self.social_acct.user = connect_user
            else:  # if auto signup enabled, we perform a quick user creation.
                self.user.set_unusable_password()
                self.user.save()
                self.social_acct.user = self.user
            self.social_acct.save()
            return self.social_acct

    def serialize(self) -> dict:
        def default_social_acct_serializer(social_acct:SocialAccount) -> dict:
            return {'uid': social_acct.uid, 'provider': social_acct.provider, 'extra_data': social_acct.extra_data }
        social_acct_serializer = getattr(SocialAccount, 'serializer', default_social_acct_serializer)
        ret = dict(
            account=social_acct_serializer(self.social_acct),
            user={'email':self.user.email, 'first_name':self.user.first_name, 'last_name':self.user.last_name,},
        )
        return ret

    @classmethod
    def deserialize(cls, data: dict):
        # todo: redit
        social_acct_deserializer = getattr(SocialAccount, 'deserializer',
                                           lambda acct: general_deserializer(SocialAccount, acct))
        social_acct = social_acct_deserializer(data["account"])
        user = User(email=data["user"]['email'], first_name=data["user"]['first_name'], last_name=data["user"]['last_name'])
        ret = cls()
        ret.social_acct = social_acct
        ret.user = user
        return ret


deserialize_social_acct = OauthAdapter.deserialize



from django.apps import AppConfig


class AuthFrameworkConfig(AppConfig):
    name = "auth_framework"
    verbose_name = "Django Auth Framework"

    def ready(self):
        import auth_framework.signals.handler


from django.conf import settings
from django.core.mail import EmailMessage
from django.template import TemplateDoesNotExist
from django.template.loader import render_to_string


def render_mail(subject: str, email: str, context: dict[str:str], template: str, **kwargs):
    """
    Renders an e-mail to `email`.  `template_prefix` identifies the
    e-mail that is to be sent, e.g. "account/email/email_confirmation"
    """
    try:
        bodies = render_to_string(template, context).strip()
    except TemplateDoesNotExist:
        raise
    msg = EmailMessage(subject, bodies,
                       kwargs.get('from_email', settings.DEFAULT_FROM_EMAIL),
                       [email])
    msg.content_subtype = 'html'  # Main content is now text/html
    return msg



__author__ = 'Carrycat'

from django.contrib.auth import get_user_model
from django.db.models.signals import pre_save, pre_delete
from django.dispatch import receiver

User = get_user_model()


@receiver(pre_delete, sender=User)
def auto_delete_avatar_on_delete(sender, instance, **kwargs):
    # Pass false so FileField doesn't save the model.
    # delete file but dont save db it will be saved later.
    if hasattr(instance, 'avatar'):
        instance.avatar.delete(False)


@receiver(pre_save, sender=User)
def auto_delete_avatar_on_change(sender, instance, update_fields, **kwargs):
    if not hasattr(instance, 'avatar'):
        return False
    if not instance.pk:
        return False

    try:
        current_instance = sender.objects.get(pk=instance.pk)
        old_image = current_instance.avatar
    except sender.DoesNotExist:
        return False
    # if not instance.image:
    #     old_image.delete(save=False)
    if (instance.avatar and not instance.avatar._committed) or not instance.avatar:
        old_image.delete(save=False)



from oauth2_provider.models import get_application_model
from rest_framework import status
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import IsAuthenticated, AllowAny
from rest_framework.response import Response

from .providers.apple.adapter import AppleOauthAdapter
from .providers.facebook.adapter import FacebookOauthAdapter
from .providers.google.adapter import GoogleOauthAdapter
from .serializers import SocialConnectionSerializer, SocialLoginSerializer
from ..oauth.serializers import AccessTokenSerializer
from ..oauth.utils import get_access_token

Application = get_application_model()


class SocialLoginView(GenericAPIView):
    permission_classes = (AllowAny,)
    serializer_class = SocialLoginSerializer
    response_serializer = AccessTokenSerializer
    process = 'login'

    def validate_client(self, raise_exception=False):
        # TODO: less confidential skip validate client(validate_client_id) for performance
        _errors = ""
        client_id = self.request.data.get('client_id')
        if not client_id:
            client = Application.objects.get(pk=1)
        else:
            client = Application.objects.get(client_id=client_id)
        if _errors and raise_exception:
            raise AuthenticationFailed(_errors)
        self.request.client = client
        self.request.client_id = client.client_id
        return client

    def perform_login(self, user):
        self.validate_client(raise_exception=True)
        token = get_access_token(user, request=self.request)
        # login(self.request, user) TODO: here
        if 'user' in token:
            token['user']['linked_social'] = [sc.provider for sc in user.socialaccount_set.all()]
        return token

    def post(self, request, *args, **kwargs):
        process = request.GET.get('process', self.process)
        serializer = self.get_serializer(data=self.request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data['user']
        resp = {}
        if process == 'login':
            resp = self.perform_login(user)
        else:
            resp['linked_social'] = [sc.provider for sc in self.request.user.socialaccount_set.all()]
        return Response(resp, status=status.HTTP_200_OK)


class SocialConnections(GenericAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = SocialConnectionSerializer

    def get(self, request):
        data = {'linked_social': [sc.provider for sc in request.user.socialaccount_set.all()]}
        return Response(data)

    def post(self, request):
        data = request.data
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({'message': 'success'})


class GoogleLogin(SocialLoginView):
    adapter_class = GoogleOauthAdapter


class FacebookLogin(SocialLoginView):
    adapter_class = FacebookOauthAdapter


class AppleLogin(SocialLoginView):
    adapter_class = AppleOauthAdapter

